"""A module for the Molecular Oncology Almanac harvester"""
import logging
from typing import Optional

import requests
import requests_cache

from metakb.harvesters.base import Harvester


logger = logging.getLogger('metakb.harvesters.moa')
logger.setLevel(logging.DEBUG)


class MOAHarvester(Harvester):
    """A class for the Molecular Oncology Almanac harvester."""

    def harvest(self, filename: Optional[str] = None):
        """
        Retrieve and store sources, variants, and assertions
        records from MOAlmanac in composite and individual JSON files.

        :param Optional[str] filename: File name for composite json
        :return: True if successfully retreived, False otherwise
        :rtype: bool
        """
        try:
            assertion_resp = self._get_all_assertions()
            sources = self._harvest_sources(assertion_resp)
            variants, variants_list = self._harvest_variants()
            assertions = \
                self._harvest_assertions(assertion_resp, variants_list)
            json_created = self.create_json(
                {
                    "assertions": assertions,
                    "sources": sources,
                    "variants": variants
                },
                filename
            )
            if not json_created:
                logger.error("MOAlmanac Harvester was not successful.")
                return False
        except Exception as e:  # noqa: E722
            logger.error(f'MOAlmanac Harvester was not successful: {e}')
            return False
        else:
            logger.info('MOAlmanac Harvester was successful.')
            return True

    def _harvest_sources(self, assertion_resp):
        """
        Harvest all MOA sources

        :param: A list of MOA assertion records
        :return: A list of sources
        :rtype: list
        """
        sources = []

        for assertion in assertion_resp:
            source = assertion['sources'][0]
            s = self._source_item(source)
            if s not in sources:
                sources.append(s)

        return sources

    def _harvest_variants(self):
        """
        Harvest all MOA variants

        :return: A list of variants
        :rtype: list
        """
        variants_list = self._get_all_variants()
        variants = []

        for variant in variants_list:
            v = self._harvest_variant(variant)
            variants.append(v)

        return variants, variants_list

    def _harvest_assertions(self, assertion_resp, variants_list):
        """
        Harvest all MOA assertions

        :param: A list of MOA assertion records
        :param: A list of MOA variant records
        :return: A list of assertions
        :rtype: list
        """
        assertions = []
        for assertion in assertion_resp:
            a = self._harvest_assertion(assertion, variants_list)
            assertions.append(a)

        return assertions

    def _get_all_assertions(self):
        """
        Return all assertion records.

        :return: All moa assertion records
        """
        with requests_cache.disabled():
            r = requests.get('https://moalmanac.org/api/assertions')
            assertions = r.json()

        return assertions

    def _get_all_variants(self):
        """
        Return all variant records

        :return: All moa variant records
        """
        with requests_cache.disabled():
            r = requests.get('https://moalmanac.org/api/features')
            variants = r.json()

        return variants

    def _source_item(self, source):
        """
        Harvest an individual MOA source of evidence

        :param: source record of each assertion record
        :return: a dictionary containing MOA source of evidence data
        :rtype: dict
        """
        source_record = {
            'id': source['source_id'],
            'type': source['source_type'],
            'doi': source['doi'],
            'nct': source['nct'],
            'pmid': source['pmid'],
            'url': source['url'],
            'citation': source['citation']
        }
        return source_record

    def _harvest_variant(self, variant):
        """
        Harvest an individual MOA variant record.

        :param: A MOA variant record
        :return: A dictionary containing MOA variant data
        :rtype: dict
        """
        variant_record = {
            'id': variant['feature_id']
        }

        variant_record.update({k: v for k, v in variant['attributes'][0].items()})  # noqa: E501
        variant_record.update(self._get_feature(variant_record))

        return variant_record

    def _harvest_assertion(self, assertion, variants_list):
        """
        Harvest an individual MOA assertion record

        :param: a MOA assertion record
        :param: a list of MOA variant records
        :return: A dictionary containing MOA assertion data
        :rtype: dict
        """
        assertion_record = {
            'id': assertion['assertion_id'],
            'context': assertion['context'],
            'description': assertion['description'],
            'disease': {
                'name': assertion['disease'],
                'oncotree_code': assertion['oncotree_code'],
                'oncotree_term': assertion['oncotree_term']
            },
            'therapy_name': assertion['therapy_name'],
            'therapy_type': assertion['therapy_type'],
            'clinical_significance': self._get_therapy(
                assertion['therapy_resistance'],
                assertion['therapy_sensitivity']),
            'predictive_implication': assertion["predictive_implication"],
            'favorable_prognosis': assertion['favorable_prognosis'],
            'created_on': assertion['created_on'],
            'last_updated': assertion['last_updated'],
            'submitted_by': assertion['submitted_by'],
            'validated': assertion['validated'],
            'source_ids': assertion['sources'][0]['source_id']
        }

        for v in variants_list:
            if v['attributes'][0] == assertion['features'][0]['attributes'][0]:
                assertion_record.update({'variant': self._harvest_variant(v)})

        return assertion_record

    def _get_therapy(self, resistance, sensitivity):
        """
        Get therapy response data.

        :param: therapy_resistance
        :param: therapy_sensitivity
        :return: whether the therapy response is resistance or sensitivity
        :rtype: str
        """
        if resistance:
            return "resistance"
        elif sensitivity:
            return "sensitivity"
        else:
            return

    def _get_feature(self, v):
        """
        Get feature name from the harvested variants

        :param: harvested MOA variant
        :return: feature name same format as displayed in moalmanac.org
        :rtype: dict
        """
        feature_type = v['feature_type']
        if feature_type == 'rearrangement':
            feature = '{}{}{}'.format(v['gene1'],
                                      f"--{v['gene2']}" if v['gene2'] else '',
                                      f" {v['rearrangement_type']}"
                                      if v['rearrangement_type'] else '')
        elif feature_type == 'somatic_variant':
            feature = '{}{}{}'.format(v['gene'],
                                      f" {v['protein_change']}"
                                      if v['protein_change'] else '',
                                      f" ({v['variant_annotation']})"
                                      if v['variant_annotation'] else '')
        elif feature_type == 'germline_variant':
            feature = '{}{}'.format(v['gene'], ' (Pathogenic)'
                                    if v['pathogenic'] == '1.0' else '')
        elif feature_type == 'copy_number':
            feature = '{} {}'.format(v['gene'], v['direction'])
        elif feature_type == 'microsatellite_stability':
            feature = '{}'.format(v.get('status'))
        elif feature_type == 'mutational_signature':
            csn = v['cosmic_signature_number']
            feature = 'COSMIC Signature {}'.format(csn)
        elif feature_type == 'mutational_burden':
            clss = v['classification']
            min_mut = v['minimum_mutations']
            mut_per_mb = v['mutations_per_mb']
            feature = '{}{}'.format(clss,
                                    f" (>= {min_mut} mutations)" if min_mut
                                    else (f" (>= {mut_per_mb} mutations/Mb)"
                                          if mut_per_mb else ''))
        elif feature_type == 'neoantigen_burden':
            feature = '{}'.format(v['classification'])
        elif feature_type == 'knockdown' or feature_type == 'silencing':
            feature = '{}{}'.format(v['gene'], f" ({v['technique']})"
                                    if v['technique'] else '')
        else:
            feature = '{}'.format(v['event'])

        return {'feature': feature.strip()}



"""A module to convert MOA resources to common data model"""
from typing import Optional
import logging
from urllib.parse import quote

from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor,\
    Extension, GeneDescriptor, ValueObjectDescriptor

import metakb.schemas as schemas
from metakb.transform.base import Transform

logger = logging.getLogger('metakb.transform.moa')
logger.setLevel(logging.DEBUG)


class MOATransform(Transform):
    """A class for transforming MOA resources to common data model."""

    def transform(self):
        """Transform MOA harvested JSON to common date model.
        Saves output in MOA transform directory.
        """
        data = self.extract_harvester()
        cdm_assertions = {}  # assertions that have been transformed to CDM

        assertions = data['assertions']
        sources = data['sources']
        variants = data['variants']

        # Transform MOA assertions
        self._transform_statements(assertions, variants, sources,
                                   cdm_assertions)

    def _transform_statements(self, records, variants, sources,
                              cdm_assertions):
        """Add transformed assertions to the response list.

        :param: A list of MOA assertion records
        :param: A dict of MOA variant records
        :param: A dict of MOA source records
        :param: Keeps track of proposition and support_evidence indexes
        :param: A dict containing assertions that have been
            transformed to the CDM
        """
        for record in records:
            gene_descriptors = self._get_gene_descriptors(
                self._get_record(record['variant']['id'], variants))
            descriptors = \
                self._get_descriptors(record, variants, gene_descriptors)
            if not descriptors:
                continue
            else:
                therapy_descriptors, variation_descriptors, disease_descriptors = descriptors  # noqa: E501

            propositions = \
                self._get_tr_propositions(record, variation_descriptors,
                                          disease_descriptors,
                                          therapy_descriptors)

            # We only want therapeutic response for now
            if not propositions:
                continue

            documents = self._get_documents(
                self._get_record(record['source_ids'], sources))

            methods = self._get_method()
            statements = self._get_statement(record, propositions,
                                             variation_descriptors,
                                             therapy_descriptors,
                                             disease_descriptors,
                                             methods, documents)

            response = schemas.Response(
                statements=statements,
                propositions=propositions,
                variation_descriptors=variation_descriptors,
                gene_descriptors=gene_descriptors,
                therapy_descriptors=therapy_descriptors,
                disease_descriptors=disease_descriptors,
                methods=methods,
                documents=documents
            ).dict(by_alias=True, exclude_none=True)

            cdm_assertions[f"moa:assertion_{record['id']}"] = response

            for field in ['statements', 'propositions',
                          'variation_descriptors', 'gene_descriptors',
                          'therapy_descriptors', 'disease_descriptors',
                          'methods', 'documents']:
                attr = getattr(self, field)
                var = response[field]
                for el in var:
                    if el not in attr:
                        attr.append(el)

    def _get_descriptors(self, record, variants, gene_descriptors):
        """Return tuple of descriptors if one exists for each type.

        :param: A MOA assertion record
        :param: MOA variant records
        :param: The corresponding gene descriptors
        :return: a tuple Descriptors
        """
        therapy_descriptors = self._get_therapy_descriptors(record)
        len_td = len(therapy_descriptors)
        if len_td != 1:
            logger.warning(f"Expected 1 therapy_descriptor for"
                           f" {record['therapy_name']} but found {len_td}")
            return None

        variation_descriptors = self._get_variation_descriptors(
            self._get_record(record['variant']['id'], variants),
            gene_descriptors)
        len_vd = len(variation_descriptors)
        if len_vd != 1:
            logger.warning(f"Expected 1 variation descriptor for"
                           f" {record['variant']} but found {len_vd}")
            return None

        disease_descriptors = self._get_disease_descriptors(record)
        len_dd = len(disease_descriptors)
        if len_dd != 1:
            logger.warning(f"Expected 1 disease descriptor for"
                           f" {record['disease']} but found {len_dd}")
            return None

        return therapy_descriptors, variation_descriptors, disease_descriptors

    def _get_statement(self, record, propositions, variant_descriptors,
                       therapy_descriptors, disease_descriptors,
                       methods, documents):
        """Get a statement for an assertion.

        :param dict record: A MOA assertion record
        :param list propositions: Propositions for the record
        :param list variant_descriptors: Variant Descriptors for the record
        :param list therapy_descriptors: Therapy Descriptors for the record
        :param list disease_descriptors: Disease Descriptors for the record
        :param list methods: Assertion methods for the record
        :param list documents: Supporting evidence for the rcord
        :return: A list of statement
        """
        evidence_level = record['predictive_implication'].strip().replace(' ', '_')  # noqa: E501

        statement = schemas.Statement(
            id=f"{schemas.SourceName.MOA.value}.assertion:{record['id']}",
            description=record['description'],
            evidence_level=f"moa.evidence_level:"
                           f"{evidence_level}",
            proposition=propositions[0]['id'],
            variation_origin=self._get_variation_origin(record['variant']),
            variation_descriptor=variant_descriptors[0]['id'],
            therapy_descriptor=therapy_descriptors[0]['id'],
            disease_descriptor=disease_descriptors[0]['id'],
            method=methods[0]['id'],
            supported_by=[se['id'] for se in documents]
        ).dict(exclude_none=True)

        return [statement]

    def _get_tr_propositions(self, record, variation_descriptors,
                             disease_descriptors, therapy_descriptors):
        """Return a list of propositions.

        :param: MOA assertion
        :param: A list of Variation Descriptors
        :param: A list of Disease Descriptors
        :param: A list of therapy_descriptors
        :return: A list of therapeutic propositions.
        """
        predicate = self._get_predicate(record['clinical_significance'])

        # Don't support TR that has  `None`, 'N/A', or 'Unknown' predicate
        if not predicate:
            return []

        params = {
            'id': '',
            'type': schemas.PropositionType.PREDICTIVE,
            'predicate': predicate,
            'subject': variation_descriptors[0]['variation_id'],
            'object_qualifier': disease_descriptors[0]['disease_id'],
            'object': therapy_descriptors[0]['therapy_id']
        }

        # Get corresponding id for proposition
        params["id"] = self._get_proposition_id(
            params["type"],
            params["predicate"],
            variation_ids=[params["subject"]],
            disease_ids=[params["object_qualifier"]],
            therapy_ids=[params["object"]]
        )
        proposition = schemas.TherapeuticResponseProposition(
            **params).dict(exclude_none=True)
        return [proposition]

    def _get_predicate(self,
                       clin_sig) -> Optional[schemas.PredictivePredicate]:
        """Get the predicate of this record

        :param: clinical significance of the assertion
        :return: predicate if valid, None otherwise
        """
        if not clin_sig:
            return None
        try:
            return schemas.PredictivePredicate[clin_sig.upper()]
        except KeyError:
            return None

    def _get_variation_origin(self, variant):
        """Return variant origin.

        :param: A MOA variant record
        :return: A str representation of variation origin
        """
        if variant['feature_type'] == 'somatic_variant':
            origin = schemas.VariationOrigin.SOMATIC.value
        elif variant['feature_type'] == 'germline_variant':
            origin = schemas.VariationOrigin.GERMLINE.value
        else:
            origin = None

        return origin

    def _get_variation_descriptors(self, variant, g_descriptors):
        """Add variation descriptor to therapeutic response

        :param: single assertion record from MOA
        :return: list of variation descriptor
        """
        vrs_ref_allele_seq = variant['protein_change'][2] \
            if 'protein_change' in variant and variant['protein_change'] else None  # noqa: E501

        v_norm_resp = None
        # For now, the normalizer only support a.a substitution
        if g_descriptors and 'protein_change' in variant and variant['protein_change']:  # noqa: E501
            gene = g_descriptors[0]['label']
            query = f"{gene} {variant['protein_change'][2:]}"
            v_norm_resp = self.vicc_normalizers.normalize_variation([query])

            if not v_norm_resp:
                logger.warning(f"Variant Normalizer unable to normalize: "
                               f"moa.variant:{variant['id']}.")
                return []
        else:
            logger.warning(f"Variation Normalizer does not support "
                           f"moa.variant:{variant['id']}: {variant}")
            return []

        gene_context = g_descriptors[0]['id'] if g_descriptors else None

        variation_descriptor = VariationDescriptor(
            id=f"moa.variant:{variant['id']}",
            label=variant['feature'],
            variation_id=v_norm_resp.variation_id,
            variation=v_norm_resp.variation,
            gene_context=gene_context,
            vrs_ref_allele_seq=vrs_ref_allele_seq,
            extensions=self._get_variant_extensions(variant)
        ).dict(by_alias=True, exclude_none=True)
        return [variation_descriptor]

    def _get_variant_extensions(self, variant):
        """Return a list of extensions for a variant.

        :param dict variant: A MOA variant record
        :return: A list of extensions
        """
        coordinate = ['chromosome', 'start_position', 'end_position',
                      'reference_allele', 'alternate_allele',
                      'cdna_change', 'protein_change', 'exon']

        extensions = [
            Extension(
                name='moa_representative_coordinate',
                value={c: variant[c] for c in coordinate}
            ).dict(exclude_none=True)
        ]

        if variant['rsid']:
            extensions.append(
                Extension(
                    name='moa_rsid',
                    value=variant['rsid']
                ).dict(exclude_none=True)
            )
        return extensions

    def _get_gene_descriptors(self, variant):
        """Return a Gene Descriptor.

        :param: A MOA variant record
        :return: A Gene Descriptor
        """
        genes = [value for key, value in variant.items()
                 if key.startswith('gene')]
        genes = list(filter(None, genes))

        gene_descriptors = []  # for fusion protein, we would include both genes  # noqa: E501
        if genes:
            for gene in genes:
                _, normalized_gene_id = \
                    self.vicc_normalizers.normalize_gene([gene])
                if normalized_gene_id:
                    gene_descriptor = GeneDescriptor(
                        id=f"{schemas.SourceName.MOA.value}.normalize."
                           f"{schemas.NormalizerPrefix.GENE.value}:{quote(gene)}",  # noqa: E501
                        label=gene,
                        gene_id=normalized_gene_id,
                    ).dict(exclude_none=True)
                else:
                    logger.warning(f"Gene Normalizer unable to "
                                   f"normalize: {gene}")
                    gene_descriptor = {}

                gene_descriptors.append(gene_descriptor)

        return gene_descriptors

    def _get_documents(self, source):
        """Get an assertion's documents.

        :param: An evidence source
        :param: Keeps track of proposition and documents indexes
        """
        if source['pmid'] != "None":
            documents_id = f"pmid:{source['pmid']}"
        else:
            documents_id = source['url']

        xrefs = []
        if source['doi']:
            xrefs.append(f"doi:{source['doi']}")
        if source['nct'] != "None":
            xrefs.append(f"nct:{source['nct']}")

        documents = schemas.Document(
            id=documents_id,
            label=source['citation'],
            xrefs=xrefs if xrefs else None
        ).dict(exclude_none=True)

        return [documents]

    def _get_method(self):
        """Get methods for a given record.

        :return: A list of methods
        """
        methods = [schemas.Method(
            id=f'method:'
               f'{schemas.MethodID.MOA_ASSERTION_BIORXIV}',
            label='Clinical interpretation of integrative molecular profiles to guide precision cancer medicine',  # noqa:E501
            url='https://www.biorxiv.org/content/10.1101/2020.09.22.308833v1',  # noqa:E501
            version=schemas.Date(year=2020, month=9, day=22),
            authors='Reardon, B., Moore, N.D., Moore, N. et al.'
        ).dict()]

        return methods

    def _get_therapy_descriptors(self, assertion):
        """Return a list of Therapy Descriptors.

        :param: an MOA assertion record
        :return: A list of Therapy Descriptors
        """
        label = assertion['therapy_name']

        if not label:
            return []

        therapy_norm_resp, normalized_therapy_id = \
            self.vicc_normalizers.normalize_therapy([label])

        if not normalized_therapy_id:
            logger.warning(f"Therapy Normalizer unable to normalize: {label}")
            return []

        if normalized_therapy_id:
            therapy_descriptor = ValueObjectDescriptor(
                id=f"{schemas.SourceName.MOA.value}."
                   f"{therapy_norm_resp.therapy_descriptor.id}",
                type="TherapyDescriptor",
                label=label,
                therapy_id=normalized_therapy_id
            ).dict(exclude_none=True)
        else:
            return []

        return [therapy_descriptor]

    def _get_disease_descriptors(self, assertion):
        """Return A list of Disease Descriptors.

        :param: an MOA assertion record
        :return: A list of Therapy Descriptors
        """
        ot_code = assertion['disease']['oncotree_code']
        if ot_code:
            ot_code = f"oncotree:{ot_code}"
        disease_name = assertion['disease']['name']

        disease_norm_resp, normalized_disease_id = \
            self.vicc_normalizers.normalize_disease([ot_code, disease_name])

        if not normalized_disease_id:
            logger.warning(f"Disease Normalize unable to normalize: "
                           f"{ot_code} and {disease_name}")
            return []

        disease_descriptor = ValueObjectDescriptor(
            id=f"{schemas.SourceName.MOA.value}."
               f"{disease_norm_resp.disease_descriptor.id}",
            type="DiseaseDescriptor",
            label=disease_name,
            disease_id=normalized_disease_id,
        ).dict(exclude_none=True)

        return [disease_descriptor]

    def _get_record(self, record_id, records):
        """Get a MOA record by ID.

        :param: The ID of the record we are searching for
        :param: A dict of records for a given MOA record type
        """
        for r in records:
            if r['id'] == record_id:
                return r



"""Module for queries."""
from typing import Dict, List, Optional, Tuple
import logging
import json
from json.decoder import JSONDecodeError
from urllib.parse import quote

from ga4gh.vrsatile.pydantic.vrsatile_models import Extension, Expression
from neo4j.graph import Node
from neo4j.data import Record
from neo4j.work.transaction import Transaction
from neo4j.work.simple import Session

from metakb.database import Graph
from metakb.normalizers import VICCNormalizers
from metakb.schemas import SearchService, StatementResponse, \
    TherapeuticResponseProposition, VariationDescriptor, \
    ValueObjectDescriptor, GeneDescriptor, Method, \
    Document, SearchIDService, DiagnosticProposition, PrognosticProposition, \
    SearchStatementsService, NestedStatementResponse, PropositionType, \
    Proposition, ServiceMeta, Predicate

logger = logging.getLogger("metakb.query")
logger.setLevel(logging.DEBUG)


class QueryHandler:
    """Class for handling queries."""

    def __init__(self, uri: str = "",
                 creds: Tuple[str, str] = ("", ""),
                 normalizers: VICCNormalizers = VICCNormalizers()) -> None:
        """Initialize neo4j driver and the VICC normalizers.
        :param str uri: address of Neo4j DB
        :param Tuple[str, str] credentials: tuple containing username and
            password
        :param VICCNormalizers normalizers: normalizer collection instance
        """
        self.driver = Graph(uri, creds).driver
        self.vicc_normalizers = normalizers

    def get_normalized_therapy(self, therapy: str,
                               warnings: List[str]) -> Optional[str]:
        """Get normalized therapy concept.

        :param str therapy: Therapy query
        :param List[str] warnings: A list of warnings for the search query
        :return: A normalized therapy concept if it exists
        """
        _, normalized_therapy_id = \
            self.vicc_normalizers.normalize_therapy([therapy])

        if not normalized_therapy_id:
            warnings.append(f"Therapy Normalizer unable to normalize: "
                            f"{therapy}")
        return normalized_therapy_id

    def get_normalized_disease(self, disease: str,
                               warnings: List[str]) -> Optional[str]:
        """Get normalized disease concept.

        :param str disease: Disease query
        :param List[str] warnings: A list of warnings for the search query
        :return: A normalized disease concept if it exists
        """
        _, normalized_disease_id = \
            self.vicc_normalizers.normalize_disease([disease])

        if not normalized_disease_id:
            warnings.append(f"Disease Normalizer unable to normalize: "
                            f"{disease}")
        return normalized_disease_id

    def get_normalized_variation(self, variation: str,
                                 warnings: List[str]) -> Optional[str]:
        """Get normalized variation concept.

        :param str variation: Variation query
        :param List[str] warnings: A list of warnings for the search query
        :return: A normalized variant concept if it exists
        """
        variant_norm_resp = \
            self.vicc_normalizers.normalize_variation([variation])
        normalized_variation = None
        if variant_norm_resp:
            normalized_variation = variant_norm_resp.variation_id
        if not normalized_variation:
            # Check if VRS variation (allele, cnv, or haplotype)
            if variation.startswith(("ga4gh:VA.", "ga4gh:CNV.", "ga4gh:VH.")):
                normalized_variation = variation
            else:
                warnings.append(f"Variant Normalizer unable to normalize: "
                                f"{variation}")
        return normalized_variation

    def get_normalized_gene(self, gene: str,
                            warnings: List[str]) -> Optional[str]:
        """Get normalized gene concept.

        :param str gene: Gene query
        :param List[str] warnings: A list of warnings for the search query.
        :return: A normalized gene concept if it exists
        """
        _, normalized_gene_id = self.vicc_normalizers.normalize_gene([gene])
        if not normalized_gene_id:
            warnings.append(f"Gene Normalizer unable to normalize: {gene}")
        return normalized_gene_id

    def get_normalized_terms(
        self, variation: Optional[str], disease: Optional[str],
        therapy: Optional[str], gene: Optional[str],
        statement_id: Optional[str], response: Dict
    ) -> Optional[Tuple]:
        """Find normalized terms for queried concepts.

        :param Optional[str] variation: Variation (subject) query
        :param Optional[str] disease: Disease (object_qualifier) query
        :param Optional[str] therapy: Therapy (object) query
        :param Optional[str] gene: Gene query
        :param Optional[str] statement_id: Statement ID query
        :param Dict response: The response for the query
        :return: A tuple containing the normalized concepts
        """
        if not (variation or disease or therapy or gene or statement_id):
            response["warnings"].append("No parameters were entered.")
            return None

        # Find normalized terms using VICC normalizers
        if therapy:
            response["query"]["therapy"] = therapy
            normalized_therapy = \
                self.get_normalized_therapy(therapy.strip(),
                                            response["warnings"])
        else:
            normalized_therapy = None
        if disease:
            response["query"]["disease"] = disease
            normalized_disease = \
                self.get_normalized_disease(disease.strip(),
                                            response["warnings"])
        else:
            normalized_disease = None
        if variation:
            response["query"]["variation"] = variation
            normalized_variation = \
                self.get_normalized_variation(variation,
                                              response["warnings"])
        else:
            normalized_variation = None
        if gene:
            response["query"]["gene"] = gene
            normalized_gene = self.get_normalized_gene(gene,
                                                       response["warnings"])
        else:
            normalized_gene = None

        # Check that queried statement_id is valid
        valid_statement_id = None
        statement = None
        if statement_id:
            response["query"]["statement_id"] = statement_id
            with self.driver.session() as session:
                statement = session.read_transaction(
                    self._get_statement_by_id, statement_id
                )
                if statement:
                    valid_statement_id = statement.get("id")
                else:
                    response["warnings"].append(
                        f"Statement: {statement_id} does not exist.")

        # If queried concept is given check that it is normalized / valid
        if (variation and not normalized_variation) or \
                (therapy and not normalized_therapy) or \
                (disease and not normalized_disease) or \
                (gene and not normalized_gene) or \
                (statement_id and not valid_statement_id):
            return None

        return (normalized_variation, normalized_disease, normalized_therapy,
                normalized_gene, statement, valid_statement_id)

    def search(self, variation: Optional[str] = None,
               disease: Optional[str] = None, therapy: Optional[str] = None,
               gene: Optional[str] = None, statement_id: Optional[str] = None,
               detail: bool = False) -> Dict:
        """Get statements and propositions from queried concepts.

        :param Optional[str] variation: Variation query
        :param Optional[str] disease: Disease query
        :param Optional[str] therapy: Therapy query
        :param Optional[str] gene: Gene query
        :param Optional[str] statement_id: Statement ID query
        :param bool detail: Whether or not to display all descriptors,
            methods, and documents
        :return: A dictionary containing the statements and propositions
            with relationships to the queried concepts
        """
        response: Dict = {
            "query": {
                "variation": None,
                "disease": None,
                "therapy": None,
                "gene": None,
                "statement_id": None,
                "detail": detail
            },
            "warnings": [],
            "matches": {
                "statements": [],
                "propositions": []
            },
            "statements": [],  # All Statements
            "propositions": [],  # All propositions
            "variation_descriptors": [],
            "gene_descriptors": [],
            "therapy_descriptors": [],
            "disease_descriptors": [],
            "methods": [],
            "documents": [],
            "service_meta_": ServiceMeta().dict()
        }

        normalized_terms = self.get_normalized_terms(
            variation, disease, therapy, gene, statement_id, response)
        if normalized_terms is None:
            return SearchService(**response).dict()
        (normalized_variation, normalized_disease,
         normalized_therapy, normalized_gene, statement,
         valid_statement_id) = normalized_terms

        session = self.driver.session()
        proposition_nodes = session.read_transaction(
            self._get_propositions, valid_statement_id, normalized_variation,
            normalized_therapy, normalized_disease, normalized_gene,
        )

        if not valid_statement_id:
            # If statement ID isn't specified, get all statements
            # related to a proposition
            statement_nodes = list()
            for p_node in proposition_nodes:
                p_id = p_node.get("id")
                if p_id not in response["matches"]["propositions"]:
                    response["matches"]["propositions"].append(p_id)
                statements = session.read_transaction(
                    self._get_statements_from_proposition, p_id
                )
                for s in statements:
                    statement_nodes.append(s)
                    s_id = s.get("id")
                    if s_id not in response["matches"]["statements"]:
                        response["matches"]["statements"].append(s_id)
        else:
            # Given Statement ID
            statement_nodes = [statement]
            s_id = statement.get("id")
            response["matches"]["statements"].append(s_id)

            for p in proposition_nodes:
                p_id = p.get("id")
                if p_id not in response["matches"]["propositions"]:
                    response["matches"]["propositions"].append(p_id)

        # Add statements found in `supported_by` to statement_nodes
        # Then add the associated proposition to proposition_nodes
        for s in statement_nodes:
            self.add_proposition_and_statement_nodes(
                session, s.get("id"), proposition_nodes, statement_nodes
            )

        if proposition_nodes and statement_nodes:
            response["statements"] = \
                self.get_statement_response(statement_nodes)
            response["propositions"] = \
                self.get_propositions_response(proposition_nodes)
        else:
            response["warnings"].append("Could not find statements "
                                        "associated with the queried"
                                        " concepts.")

        if detail:
            for s in response["statements"]:
                vd = self._get_variation_descriptor(
                    response,
                    session.read_transaction(
                        self._find_node_by_id,
                        s["variation_descriptor"]
                    )
                )
                if vd not in response["variation_descriptors"]:
                    response["variation_descriptors"].append(vd)
                if "therapy_descriptor" in s.keys():
                    td = self._get_therapy_descriptor(
                        session.read_transaction(
                            self._find_node_by_id, s["therapy_descriptor"]
                        )
                    )
                    if td not in response["therapy_descriptors"]:
                        response["therapy_descriptors"].append(td)
                else:
                    response["therapy_descriptors"] = []

                dd = self._get_disease_descriptor(
                    session.read_transaction(
                        self._find_node_by_id, s["disease_descriptor"]
                    )
                )
                if dd not in response["disease_descriptors"]:
                    response["disease_descriptors"].append(dd)

                m = self._get_method(
                    session.read_transaction(
                        self._find_node_by_id, s["method"]
                    )
                )
                if m not in response["methods"]:
                    response["methods"].append(m)

                # Sometimes CIViC AIDs have supported by statements
                # that we aren't able to transform
                sb_not_found = set()
                for sb_id in s["supported_by"]:
                    try:
                        document = self._get_document(
                            session.read_transaction(
                                self._find_node_by_id, sb_id
                            )
                        )
                        if document:
                            if document not in response["documents"]:
                                response["documents"].append(document)
                    except ValueError:
                        sb_not_found.add(sb_id)
                if sb_not_found:
                    response["warnings"].append(f"Supported by evidence not "
                                                f"yet  supported in MetaKB: "
                                                f"{sb_not_found} for "
                                                f"{s['id']}")
        else:
            response["variation_descriptors"] = None
            response["gene_descriptors"] = None
            response["disease_descriptors"] = None
            response["therapy_descriptors"] = None
            response["methods"] = None
            response["documents"] = None

        session.close()
        return SearchService(**response).dict(by_alias=True, exclude_none=True)

    def search_by_id(self, node_id: str) -> Dict:
        """Get node information given id query

        :param str node_id: Node's ID query
        :return: A dictionary containing the node content
        """
        valid_node_id = None
        response = {
            "query": node_id,
            "warnings": [],
            "service_meta_": ServiceMeta().dict()
        }

        if not node_id:
            response["warnings"].append("No parameters were entered.")
        elif node_id.strip() == "":
            response["warnings"].append("Cannot enter empty string.")
        else:
            node_id = node_id.strip()
            if "%" not in node_id and ":" in node_id:
                concept_name = quote(node_id.split(":", 1)[1])
                node_id = \
                    f"{node_id.split(':', 1)[0]}" \
                    f":{concept_name}"
            with self.driver.session() as session:
                node = session.read_transaction(
                    self._find_node_by_id, node_id
                )
                if node:
                    valid_node_id = node.get("id")
                else:
                    response["warnings"].append(f"Node: {node_id} "
                                                f"does not exist.")
        if (not node_id and not valid_node_id) or \
                (node_id and not valid_node_id):
            return SearchIDService(**response).dict(exclude_none=True)

        label, *_ = node.labels
        if label == "Statement":
            statement = self._get_statement(node)
            if statement:
                response["statement"] = statement
        elif label in ["Proposition", "TherapeuticResponse",
                       "Prognostic", "Diagnostic"]:
            proposition = self._get_proposition(node)
            if proposition:
                response["proposition"] = proposition
        elif label == "VariationDescriptor":
            response["variation_descriptor"] = \
                self._get_variation_descriptor(response, node)
        elif label == "TherapyDescriptor":
            response["therapy_descriptor"] = \
                self._get_therapy_descriptor(node)
        elif label == "DiseaseDescriptor":
            response["disease_descriptor"] = self._get_disease_descriptor(node)
        elif label == "GeneDescriptor":
            response["gene_descriptor"] = \
                self._get_gene_descriptor(node, self._get_gene_value_object(node))  # noqa: E501
        elif label == "Document":
            document = self._get_document(node)
            if document:
                response["document"] = document
        elif label == "Method":
            response["method"] = self._get_method(node)

        session.close()
        return SearchIDService(**response).dict(
            by_alias=True, exclude_none=True)

    def search_statements(
            self, variation: Optional[str] = None,
            disease: Optional[str] = None, therapy: Optional[str] = None,
            gene: Optional[str] = None, statement_id: Optional[str] = None
    ) -> Dict:
        """Get nested statements from queried concepts

        :param Optional[str] variation: Variation query
        :param Optional[str] disease: Disease query
        :param Optional[str] therapy: Therapy query
        :param Optional[str] gene: Gene query
        :param Optional[str] statement_id: Statement ID query
        :return: A dictionary containing the statements with nested
            propositions, descriptors, methods, and supported by documents
        """
        response: Dict = {
            "query": {
                "variation": None,
                "disease": None,
                "therapy": None,
                "gene": None,
                "statement_id": None
            },
            "warnings": [],
            "matches": {
                "statements": [],
                "propositions": []
            },
            "statements": [],
            "service_meta_": ServiceMeta().dict()
        }

        normalized_terms = self.get_normalized_terms(
            variation, disease, therapy, gene, statement_id, response)
        if normalized_terms is None:
            return SearchStatementsService(**response).dict()
        (normalized_variation, normalized_disease,
         normalized_therapy, normalized_gene, statement,
         valid_statement_id) = normalized_terms

        session = self.driver.session()
        statement_nodes = list()
        proposition_nodes = session.read_transaction(
            self._get_propositions, valid_statement_id, normalized_variation,
            normalized_therapy, normalized_disease, normalized_gene
        )

        proposition_cache = dict()
        if not valid_statement_id:
            # If statement ID isn"t specified, get all statements
            # related to a proposition
            for p_node in proposition_nodes:
                p_id = p_node.get("id")
                if p_id not in response["matches"]["propositions"]:
                    response["matches"]["propositions"].append(p_id)
                self._add_to_proposition_cache(
                    session, p_node, proposition_cache)
                statements = session.read_transaction(
                    self._get_statements_from_proposition, p_node.get("id")
                )
                for s in statements:
                    statement_nodes.append(s)
                    s_id = s.get("id")
                    if s_id not in response["matches"]["statements"]:
                        response["matches"]["statements"].append(s_id)
        else:
            # Given Statement ID
            statement_nodes.append(statement)
            s_id = statement.get("id")
            p_node = proposition_nodes[0]
            p_id = p_node.get("id")
            if s_id not in response["matches"]["statements"]:
                response["matches"]["statements"].append(statement_id)
            if p_id not in response["matches"]["propositions"]:
                response["matches"]["propositions"].append(p_id)
            self._add_to_proposition_cache(session, p_node, proposition_cache)

        # Add statements found in `supported_by` to statement_nodes
        # Then add the associated proposition to proposition_nodes
        og_prop_nodes_len = len(proposition_nodes)
        for s in statement_nodes:
            self.add_proposition_and_statement_nodes(
                session, s.get("id"), proposition_nodes, statement_nodes
            )

            if og_prop_nodes_len != len(proposition_nodes):
                for p_node in proposition_nodes:
                    self._add_to_proposition_cache(
                        session, p_node, proposition_cache)

        methods_cache: Dict = dict()
        variations_cache: Dict = dict()
        disease_cache: Dict = dict()
        therapy_cache: Dict = dict()
        document_cache: Dict = dict()
        added_statements = set()

        for s in statement_nodes:
            s_id = s.get("id")
            if s_id in added_statements:
                continue

            statement_resp = session.read_transaction(
                self._find_and_return_statement_response, s_id
            )
            p_id = statement_resp.get("p_id")
            proposition = proposition_cache[p_id]

            method_id = statement_resp["m"]["id"]
            if method_id in methods_cache:
                method = methods_cache[method_id]
            else:
                method = self.search_by_id(method_id)["method"]
                methods_cache[method_id] = method

            variation_id = statement_resp["vid"]
            if variation_id in variations_cache:
                variation_descr = variations_cache[variation_id]
            else:
                variation_descr = self._get_variation_descriptor(
                    {},
                    session.read_transaction(
                        self._find_node_by_id, variation_id),
                    gene_context_by_id=False
                )
                variations_cache[variation_id] = variation_descr

            if proposition.type == PropositionType.PREDICTIVE:
                therapy_id = statement_resp.get("tid")
                if therapy_id in therapy_cache:
                    therapy_descr = therapy_cache[therapy_id]
                else:
                    therapy_descr = self._get_therapy_descriptor(
                        session.read_transaction(self._find_node_by_id,
                                                 therapy_id)
                    )
                    therapy_cache[therapy_id] = therapy_descr
            else:
                therapy_descr = None

            disease_id = statement_resp.get("did")
            if disease_id in disease_cache:
                disease_descr = disease_cache[disease_id]
            else:
                disease_descr = self._get_disease_descriptor(
                    session.read_transaction(self._find_node_by_id,
                                             disease_id)
                )
                disease_cache[disease_id] = disease_descr

            supported_by = list()
            sb_not_found = set()
            sb_list = session.read_transaction(
                self._find_and_return_supported_by, s_id
            )
            for sb in sb_list:
                sb_id = sb.get("id")
                try:
                    if sb_id in document_cache:
                        document = document_cache[sb_id]
                    else:
                        document = self._get_document(
                            session.read_transaction(
                                self._find_node_by_id, sb_id
                            )
                        )

                    if document:
                        supported_by.append(document)
                        document_cache[sb_id] = document
                    else:
                        if sb_id.startswith("civic.eid"):
                            supported_by.append(sb_id)
                except ValueError:
                    sb_not_found.add(sb_id)
            if sb_not_found:
                response["warnings"].append(f"Supported by evidence not "
                                            f"yet  supported in MetaKB: "
                                            f"{sb_not_found} for "
                                            f"{s['id']}")

            params = {
                "id": s_id,
                "description": s.get("description"),
                "direction": s.get("direction"),
                "evidence_level": s.get("evidence_level"),
                "variation_origin": s.get("variation_origin"),
                "proposition": proposition,
                "variation_descriptor": variation_descr,
                "therapy_descriptor": therapy_descr,
                "disease_descriptor": disease_descr,
                "method": method,
                "supported_by": supported_by
            }
            response["statements"].append(
                NestedStatementResponse(**params).dict())
            added_statements.add(s_id)
        session.close()
        return SearchStatementsService(**response).dict(
            by_alias=True, exclude_none=True)

    def _add_to_proposition_cache(self, session: Session, p_node: Node,
                                  proposition_cache: Dict) -> None:
        """Add a proposition to `proposition_cache`

        :param Session session: Neo4j driver session
        :param Node p_node: Proposition node
        :param Dict proposition_cache: Proposition lookup dictionary
        """
        p_id = p_node.get("id")
        if p_id not in proposition_cache:
            proposition_resp = session.read_transaction(
                self._find_and_return_proposition_response,
                p_id
            )
            proposition_type = p_node.get("type")
            proposition = {
                "id": p_id,
                "type": proposition_type,
                "predicate": p_node.get("predicate"),
                "subject": proposition_resp["subject"],
                "object_qualifier": proposition_resp["object_qualifier"]
            }
            if proposition_type == PropositionType.PREDICTIVE:
                proposition["object"] = proposition_resp["object"]
                proposition = \
                    TherapeuticResponseProposition(**proposition)
            elif proposition_type == PropositionType.PROGNOSTIC:
                proposition = PrognosticProposition(**proposition)
            elif proposition_type == PropositionType.DIAGNOSTIC:
                proposition = DiagnosticProposition(**proposition)
            else:
                raise ValueError(f"{proposition_type} is not a valid "
                                 f"proposition type")
            if proposition:
                proposition_cache[p_id] = proposition

    def _get_variation_descriptor(
        self, response: Dict, variation_descriptor: Node,
            gene_context_by_id: bool = True) -> VariationDescriptor:
        """Get variation descriptor

        :param Dict response: Query response object
        :param variation_descriptor: Variation Descriptor Node
        :param bool gene_context_by_id: `True` if gene_context field will be
            a CURIE that reference's a gene descriptor. `False` if gene_context
            field will be a gene descriptor
        :return: Variation descriptor
        """
        keys = variation_descriptor.keys()
        vid = variation_descriptor.get("id")
        vd_params = {
            "id": vid,
            "label": variation_descriptor.get("label"),
            "description": variation_descriptor.get("description"),
            "variation_id": None,
            "variation": None,
            "gene_context": None,
            "molecule_context": variation_descriptor.get("molecule_context"),
            "structural_type": variation_descriptor.get("structural_type"),
            "vrs_ref_allele_seq": variation_descriptor.get("vrs_ref_allele_seq"),  # noqa: E501
            "expressions": [],
            "xrefs": variation_descriptor.get("xrefs"),
            "alternate_labels": variation_descriptor.get("alternate_labels"),
            "extensions": []
        }

        # Get Gene Descriptor / gene context
        with self.driver.session() as session:
            gene_descriptor = session.read_transaction(
                self._get_variation_descriptors_gene, vd_params["id"]
            )
            gene_descriptor_id = gene_descriptor.get("id")

            gene_value_object = session.read_transaction(
                self._find_descriptor_value_object,
                gene_descriptor_id
            )
            gene_context = self._get_gene_descriptor(
                gene_descriptor, gene_value_object)

            if gene_context_by_id:
                # Reference gene descriptor by id
                vd_params["gene_context"] = gene_descriptor_id
            else:
                # gene context will be gene descriptor
                vd_params["gene_context"] = gene_context

            if "gene_descriptors" in response and\
                    gene_descriptor_id not in response["gene_descriptors"]:
                response["gene_descriptors"].append(gene_context)

        # Get Variation Descriptor Expressions
        for key in ["expressions_genomic", "expressions_protein",
                    "expressions_transcript"]:
            if key in keys:
                for value in variation_descriptor.get(key):
                    vd_params["expressions"].append(
                        Expression(
                            syntax=f"hgvs:{key.split('_')[-1]}",
                            value=value
                        ).dict()
                    )
        if not vd_params["expressions"]:
            del vd_params["expressions"]

        # Get Variation Descriptor Extensions
        if vd_params["id"].startswith("civic.vid"):
            for field in ["civic_representative_coordinate",
                          "civic_actionability_score"]:
                if field in keys:
                    vd_params["extensions"].append(
                        Extension(
                            name=field,
                            value=json.loads(variation_descriptor.get(field))
                        ).dict()
                    )
            with self.driver.session() as session:
                variant_group = session.read_transaction(
                    self._get_variation_group, vid
                )
                if variant_group:
                    variant_group = variant_group[0]
                    vg = Extension(
                        name="variant_group",
                        value=[{
                            "id": variant_group.get("id"),
                            "label": variant_group.get("label"),
                            "description": variant_group.get("description"),
                            "type": "variant_group"
                        }]
                    ).dict()
                    for v in vg["value"]:
                        if not v["description"]:
                            del v["description"]
                    vd_params["extensions"].append(vg)
        elif vd_params["id"].startswith("moa.variant"):
            for field in ["moa_representative_coordinate", "moa_rsid"]:
                if field in keys:
                    vd_params["extensions"].append(
                        Extension(
                            name=field,
                            value=json.loads(variation_descriptor.get(field))
                        ).dict()
                    )

        with self.driver.session() as session:
            value_object = session.read_transaction(
                self._find_descriptor_value_object, vd_params["id"]
            )
            vd_params["variation_id"] = value_object.get("id")
            vd_params["variation"] = json.loads(value_object["variation"])
        return VariationDescriptor(**vd_params)

    @staticmethod
    def _get_variation_group(tx: Transaction, vid: str) -> Optional[Record]:
        """Get a variation descriptor's variation group.

        :param Transaction tx: Neo4j session transaction
        :param str vid: variation descriptor ID
        :return: query record, containing variation group node if successful
        """
        query = (
            "MATCH (vd:VariationDescriptor)-[:IN_VARIATION_GROUP]->(vg:VariationGroup) "  # noqa: E501
            f"WHERE toLower(vd.id) = toLower('{vid}') "
            "RETURN vg"
        )
        return tx.run(query).single()

    @staticmethod
    def _get_variation_descriptors_gene(tx: Transaction,
                                        vid: str) -> Optional[Node]:
        """Get a Variation Descriptor's Gene Descriptor.
        :param Transaction tx: Neo4j session transaction
        :param str vid: variation descriptor ID
        :return: Gene descriptor Node if successful
        """
        query = (
            "MATCH (vd:VariationDescriptor)-[:HAS_GENE]->(gd:GeneDescriptor) "
            f"WHERE toLower(vd.id) = toLower('{vid}') "
            "RETURN gd"
        )
        return tx.run(query).single()[0]

    @staticmethod
    def _get_gene_descriptor(gene_descriptor: Node,
                             gene_value_object: Node) -> GeneDescriptor:
        """Add gene descriptor to response.

        :param Node gene_descriptor: Gene Descriptor Node
        :param Node gene_value_object: Gene Node
        :return: GeneDescriptor object
        """
        gd_params = {
            "id": gene_descriptor.get("id"),
            "type": "GeneDescriptor",
            "label": gene_descriptor.get("label"),
            "description": gene_descriptor.get("description"),
            "gene_id": gene_value_object.get("id"),
            "alternate_labels": gene_descriptor.get("alternate_labels"),
            "xrefs": gene_descriptor.get("xrefs")
        }

        return GeneDescriptor(**gd_params)

    def _get_therapy_descriptor(
        self, therapy_descriptor: Node
    ) -> ValueObjectDescriptor:
        """Get therapy descriptor.

        :param Node therapy_descriptor: Therapy Descriptor Node
        :return: Value Object Descriptor for therapy
        """
        td_params = {
            "id": therapy_descriptor.get("id"),
            "type": "TherapyDescriptor",
            "label": therapy_descriptor.get("label"),
            "therapy_id": None,
            "alternate_labels": therapy_descriptor.get("alternate_labels"),
            "xrefs": therapy_descriptor.get("xrefs")
        }

        with self.driver.session() as session:
            value_object = session.read_transaction(
                self._find_descriptor_value_object, td_params["id"]
            )
            td_params["therapy_id"] = value_object.get("id")

        return ValueObjectDescriptor(**td_params)

    def _get_disease_descriptor(
        self, disease_descriptor: Node
    ) -> ValueObjectDescriptor:
        """Get disease descriptor.

        :param Node disease_descriptor: Disease Descriptor Node
        :return: Value Object Descriptor for disease
        """
        dd_params = {
            "id": disease_descriptor.get("id"),
            "type": "DiseaseDescriptor",
            "label": disease_descriptor.get("label"),
            "disease_id": None,
            "xrefs": disease_descriptor.get("xrefs")
        }

        with self.driver.session() as session:
            value_object = session.read_transaction(
                self._find_descriptor_value_object, dd_params["id"]
            )
            dd_params["disease_id"] = value_object.get("id")

        return ValueObjectDescriptor(**dd_params)

    @staticmethod
    def _get_method(method: Node) -> Method:
        """Get method

        :param Node method: Method Node
        :return: Method
        """
        params = dict()
        for key in method.keys():
            try:
                params[key] = json.loads(method.get(key))
            except JSONDecodeError:
                params[key] = method.get(key)

        return Method(**params)

    @staticmethod
    def _get_document(document: Node) -> Optional[Document]:
        """Add document to response.

        :param Node document: Document Node
        :return: Document if node has type `Document`
        """
        label, *_ = document.labels
        if label != "Document":
            return None

        params = dict()
        for key in document.keys():
            params[key] = document.get(key)
        return Document(**params)

    @staticmethod
    def _find_node_by_id(tx: Transaction, node_id: str) -> Optional[Node]:
        """Find a node by its ID.
        :param Transaction tx: Neo4j session transaction object
        :param str node_id: ID of node to retrieve
        :return: Node object if successful
        """
        query = (
            "MATCH (n) "
            f"WHERE toLower(n.id) = toLower('{node_id}') "
            "RETURN n"
        )
        return (tx.run(query).single() or [None])[0]

    @staticmethod
    def _find_descriptor_value_object(tx: Transaction,
                                      descriptor_id: str) -> Optional[Node]:
        """Find a Descriptor's value object.
        :param Transaction tx: Neo4j session transaction object
        :param str descriptor_id: ID of descriptor to look up
        :return: Node of value object described by descriptor if successful
        """
        query = (
            "MATCH (d)-[:DESCRIBES]->(v)"
            f"WHERE toLower(d.id) = toLower('{descriptor_id}') "
            "RETURN v"
        )
        return tx.run(query).single()[0]

    def add_proposition_and_statement_nodes(
            self, session, statement_id: str, proposition_nodes: List,
            statement_nodes: List):
        """Get statements found in `supported_by` and their propositions
        and add to corresponding list.

        :param session: Session
        :param str statement_id: Statement ID
        :param List proposition_nodes: List of propositions
        :param List statement_nodes: List of statements
        """
        supported_by_statements = session.read_transaction(
            self._find_and_return_supported_by, statement_id,
            only_statement=True
        )
        for s in supported_by_statements:
            if s not in statement_nodes:
                statement_nodes.append(s)
                proposition = session.read_transaction(
                    self._find_and_return_propositions_from_statement,
                    s.get("id")
                )
                if proposition and proposition \
                        not in proposition_nodes:
                    proposition_nodes.append(proposition)

    @staticmethod
    def _get_statement_by_id(tx: Transaction,
                             statement_id: str) -> Optional[Node]:
        """Get a Statement node by ID.

        :param Transaction tx: Neo4j session transaction object
        :param str statement_id: statemend ID to retrieve
        :return: statement node if successful
        """
        query = (
            "MATCH (s:Statement) "
            f"WHERE toLower(s.id) = toLower('{statement_id}') "
            "RETURN s"
        )
        return (tx.run(query).single() or [None])[0]

    @staticmethod
    def _get_propositions(
            tx: Transaction,
            statement_id: str = "",
            normalized_variation: str = "",
            normalized_therapy: str = "",
            normalized_disease: str = "",
            normalized_gene: str = "",
            prop_type: Optional[PropositionType] = None,
            pred: Optional[Predicate] = None
    ) -> List[Node]:
        """Get propositions that contain normalized concepts queried. Used
        as callback for Neo4j session API.

        :param Transaction tx: Neo4j session transaction object
        :param str statement_id: statement ID as stored in DB
        :param str normalized_variation: variation VRS ID
        :param str normalized_therapy: normalized therapy concept ID
        :param str normalized_disease: normalized disease concept ID
        :param str normalized_gene: normalized gene concept ID
        :param Optional[PropositionType] prop_type: type of proposition
        :param Optional[Predicate] pred: predicate value
        :returns: List of nodes matching given parameters
        """
        query = ""
        params: Dict[str, str] = {}
        if prop_type and pred:
            query += \
                "MATCH (p:Proposition {type:$prop_type, predicate:$pred}) "
            params["prop_type"] = prop_type.value
            params["pred"] = pred.value
        elif prop_type:
            query += "MATCH (p:Proposition {type:$prop_type}) "
            params["prop_type"] = prop_type.value
        elif pred:
            query += "MATCH (p:Proposition {predicate:$pred}) "
            params["pred"] = pred.value
        if statement_id:
            query += "MATCH (:Statement {id:$s_id})-[:DEFINED_BY]-> (p:Proposition) "  # noqa: E501
            params["s_id"] = statement_id
        if normalized_therapy:
            query += \
                "MATCH (p:Proposition)<-[:IS_OBJECT_OF]-(:Therapy {id:$t_id}) "
            params["t_id"] = normalized_therapy
        if normalized_variation:
            lower_normalized_variation = normalized_variation.lower()
            query += "MATCH (p:Proposition)<-[:IS_SUBJECT_OF]-(v:Variation "
            if lower_normalized_variation.startswith('ga4gh:sq.'):
                # Sequence ID
                query += "{location_sequence_id: $v_id}) "
            else:
                query += "{id:$v_id}) "
            params["v_id"] = normalized_variation
        if normalized_disease:
            query += "MATCH (p:Proposition)<-[:IS_OBJECT_QUALIFIER_OF]-(:Disease {id:$d_id}) "  # noqa: E501
            params["d_id"] = normalized_disease
        if normalized_gene:
            query += "MATCH (:Gene {id:$g_id})<-[:DESCRIBES]-" \
                     "(:GeneDescriptor)<-[:HAS_GENE]-" \
                     "(:VariationDescriptor)-[:DESCRIBES]->(v:Variation)-" \
                     "[:IS_SUBJECT_OF]->(p:Proposition) "
            params["g_id"] = normalized_gene
        query += "RETURN DISTINCT p"
        return [p[0] for p in tx.run(query, **params)]

    @staticmethod
    def _get_statements_from_proposition(tx: Transaction,
                                         proposition_id: str) -> List[Node]:
        """Get statements that are defined by a proposition.

        :param Transaction tx: Neo4j session transaction object
        :param str proposition_id: ID for proposition to retrieve associated
            statements from
        :return: List of statement Nodes
        """
        query = (
            "MATCH (p:Proposition {id: $proposition_id})<-[:DEFINED_BY]-(s:Statement) "  # noqa: E501
            "RETURN DISTINCT s"
        )
        return [s[0] for s in tx.run(query, proposition_id=proposition_id)]

    def get_statement_response(self,
                               statement_nodes: List[Node]) -> List[Dict]:
        """Return a list of statements from Statement and Proposition nodes.

        :param List statement_nodes: A list of Statement Nodes
        :return: A list of dicts containing statement response output
        """
        statements_response = list()
        added_statements = set()
        for s in statement_nodes:
            s_id = s.get("id")
            if s_id not in added_statements:
                statements_response.append(
                    self._get_statement(s)
                )
                added_statements.add(s_id)

        return statements_response

    @staticmethod
    def _find_and_return_statement_response(
        tx: Transaction, statement_id: str
    ) -> Optional[Record]:
        """Return IDs and method related to a Statement.
        :param Transaction tx: Neo4j session transaction object
        :param str statement_id: ID of statement to retrieve
        :return: Record containing descriptors, methods, and propositions
            associated with statement if successful
        """
        queries = (
            ("MATCH (s)-[r1]->(td:TherapyDescriptor) ", "td.id AS tid,"),
            ("", "")
        )
        for q in queries:
            query = (
                "MATCH (s:Statement) "
                f"WHERE s.id = '{statement_id}' "
                f"{q[0]}"
                "MATCH (s)-[r2]->(vd:VariationDescriptor) "
                "MATCH (s)-[r3]->(dd:DiseaseDescriptor) "
                "MATCH (s)-[r4]->(m:Method) "
                "MATCH (s)-[r6]->(p:Proposition) "
                f"RETURN {q[1]} vd.id AS vid, dd.id AS did, m,"
                " p.id AS p_id"
            )
            result = tx.run(query).single()
            if result:
                return result
        return None

    def get_propositions_response(
        self, proposition_nodes: List[Node]
    ) -> List[Proposition]:
        """Return a list of propositions from Proposition nodes.

        :param List[Node] proposition_nodes: A list of Proposition Nodes
        :return: A list of Propositions
        """
        propositions_response = list()
        for p in proposition_nodes:
            proposition = self._get_proposition(p)
            if proposition and proposition not in propositions_response:
                propositions_response.append(proposition)
        return propositions_response

    @staticmethod
    def _find_and_return_proposition_response(tx: Transaction,
                                              proposition_id: str) -> Record:
        """Return value ids from a proposition."""
        queries = (
            ("MATCH (n) -[r1]-> (t:Therapy) ", "t.id AS object,"), ("", "")
        )
        for q in queries:
            query = (
                f"MATCH (n) "
                f"WHERE n.id = '{proposition_id}' "
                f"{q[0]}"
                "MATCH (n) -[r2]-> (v:Variation) "
                "MATCH (n) -[r3]-> (d:Disease) "
                f"RETURN {q[1]} v.id AS subject, d.id AS object_qualifier"
            )
            result = tx.run(query).single()
            if result:
                return result
        return None

    @staticmethod
    def _find_and_return_supported_by(
        tx: Transaction, statement_id: str, only_statement: bool = False
    ) -> List[Record]:
        """Retrieve Statement and Document Nodes that support a given
        Statement.

        :param Transaction tx: Neo4j session transaction object
        :param str statement_id: ID of original statement
        :param bool only_statement: `True` if only match on Statement,
            `False` if match on both Statement and Document
        :return: List of Records that support provided statement
        """
        if not only_statement:
            match = "MATCH (s:Statement)-[:CITES]->(sb) "
        else:
            match = "MATCH (s:Statement)-[:CITES]->(sb:Statement) "
        query = (
            f"{match}"
            f"WHERE s.id = '{statement_id}' "
            "RETURN sb"
        )
        return [se[0] for se in tx.run(query)]

    @staticmethod
    def _find_and_return_propositions_from_statement(
        tx: Transaction, statement_id: str
    ) -> Optional[Node]:
        """Find propositions from a given statement.

        :param Transaction tx: Neo4j session transaction object
        :param statement_id str: statement ID to get propositions for
        :return: Node containing supported proposition if successful
        """
        query = (
            "MATCH (p:Proposition)<-[:DEFINED_BY]-(s:Statement) "
            f"WHERE toLower(s.id) = toLower('{statement_id}') "
            "RETURN p"
        )
        return (tx.run(query).single() or [None])[0]

    def _get_gene_value_object(self, node: Node) -> Node:
        """Get gene value object from gene descriptor object

        :param Node node: gene descriptor object
        :return: gene value object
        """
        with self.driver.session() as session:
            gene_value_object = session.read_transaction(
                self._find_descriptor_value_object, node.get("id")
            )
        return gene_value_object

    def _get_proposition(self, p: Node) -> Proposition:
        """Return a proposition.

        :param Node p: Proposition Node
        :return: A proposition
        :raise: ValueError if unrecognized proposition type
        """
        with self.driver.session() as session:
            p_id = p.get("id")
            p_type = p.get("type")
            proposition = None
            value_ids = session.read_transaction(
                self._find_and_return_proposition_response, p_id
            )
            params = {
                "id": p_id,
                "type": p_type,
                "predicate": p.get("predicate"),
                "subject": value_ids["subject"],
                "object_qualifier": value_ids["object_qualifier"]
            }
            if p_type == PropositionType.PREDICTIVE:
                params["object"] = value_ids["object"]
                proposition = \
                    TherapeuticResponseProposition(**params)
            elif p_type == PropositionType.PROGNOSTIC:
                proposition = PrognosticProposition(**params)
            elif p_type == PropositionType.DIAGNOSTIC:
                proposition = DiagnosticProposition(**params)
            else:
                raise ValueError
            return proposition

    def _get_statement(self, s: Node) -> Dict:
        """Return a statement.

        :param Node s: Statement Node
        :return: Dict containing values from s
        """
        with self.driver.session() as session:
            statement_id = s.get("id")
            response = session.read_transaction(
                self._find_and_return_statement_response, statement_id)
            se_list = session.read_transaction(
                self._find_and_return_supported_by, statement_id)

            statement = StatementResponse(
                id=statement_id,
                description=s.get("description"),
                direction=s.get("direction"),
                evidence_level=s.get("evidence_level"),
                variation_origin=s.get("variation_origin"),
                proposition=response["p_id"],
                variation_descriptor=response["vid"],
                therapy_descriptor=response["tid"] if "tid" in response.keys() else None,  # noqa: E501
                disease_descriptor=response["did"],
                method=response["m"]["id"],
                supported_by=[se["id"] for se in se_list]
            ).dict(exclude_none=True)
            return statement

    @staticmethod
    def _get_documents(tx: Transaction, **parameters) -> List[Node]:
        """Get Document nodes matching provided parameters. Provide as callback
        to Neo4j session methods.
        :param Transaction tx: session transaction
        :param Dict parameters: document properties and values
        :return: List of nodes matching given parameters
        :raise TypeError: if no parameters given
        """
        params_strings = []
        if len(parameters) == 0:
            raise TypeError("Must provide at least one parameter")
        for key in parameters.keys():
            params_strings.append(f"{key}:${key}")
        query_string = f"""
            MATCH (d:Document {{ {','.join(params_strings)} }})
            RETURN d
        """
        return [p[0] for p in tx.run(query_string, **parameters)]



"""The MetaKB package."""
from pathlib import Path
import logging
from os import environ

APP_ROOT = Path(__file__).resolve().parents[0]
PROJECT_ROOT = Path(__file__).resolve().parents[1]

if 'METAKB_NORM_EB_PROD' in environ:
    environ['VARIATION_NORM_EB_PROD'] = "true"
    environ['GENE_NORM_EB_PROD'] = "true"
    environ['THERAPY_NORM_EB_PROD'] = "true"
    environ['DISEASE_NORM_EB_PROD'] = "true"
    LOG_FN = "/tmp/metakb.log"
else:
    LOG_FN = "metakb.log"

logging.basicConfig(
    filename=LOG_FN,
    format='[%(asctime)s] - %(name)s - %(levelname)s : %(message)s')
logger = logging.getLogger('metakb')
logger.setLevel(logging.DEBUG)
logging.getLogger("boto3").setLevel(logging.INFO)
logging.getLogger("botocore").setLevel(logging.INFO)
logging.getLogger("urllib3").setLevel(logging.INFO)
logging.getLogger("python_jsonschema_objects").setLevel(logging.INFO)
logging.getLogger("hgvs.parser").setLevel(logging.INFO)
logging.getLogger("biocommons.seqrepo.seqaliasdb.seqaliasdb").setLevel(logging.INFO)  # noqa: E501
logging.getLogger("biocommons.seqrepo.fastadir.fastadir").setLevel(logging.INFO)  # noqa: E501
logging.getLogger("requests_cache.patcher").setLevel(logging.INFO)
logger.handlers = []

if 'METAKB_NORM_EB_PROD' in environ:
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    logger.addHandler(ch)


# default date format for all saved data files
DATE_FMT = "%Y%m%d"



"""
Provide CLI utility for performing data collection, transformation, and upload
to graph datastore.
"""
from timeit import default_timer as timer
from os import environ
import logging
from typing import Optional
from pathlib import Path

import click
from disease.database import Database as DiseaseDatabase
from disease.schemas import SourceName as DiseaseSources
from disease.cli import CLI as DiseaseCLI
from therapy.database import Database as TherapyDatabase
from therapy.schemas import SourceName as TherapySources
from therapy.cli import CLI as TherapyCLI
from gene.database import Database as GeneDatabase
from gene.schemas import SourceName as GeneSources
from gene.cli import CLI as GeneCLI

from metakb import APP_ROOT
from metakb.database import Graph
from metakb.schemas import SourceName
from metakb.harvesters import Harvester, CIViCHarvester, MOAHarvester
from metakb.transform import Transform, CIViCTransform, MOATransform


logger = logging.getLogger('metakb.cli')
logger.setLevel(logging.DEBUG)


class CLI:
    """Update database."""

    @staticmethod
    @click.command()
    @click.option(
        '--db_url',
        help=('URL endpoint for the application Neo4j database. Can also be '
              'provided via environment variable METAKB_DB_URL.')
    )
    @click.option(
        '--db_username',
        help=('Username to provide to application database. Can also be '
              'provided via environment variable METAKB_DB_USERNAME.')
    )
    @click.option(
        '--db_password',
        help=('Password to provide to application database. Can also be '
              'provided via environment variable METAKB_DB_PASSWORD.')
    )
    @click.option(
        '--load_normalizers_db',
        '-i',
        is_flag=True,
        default=False,
        help='Check normalizers database and load data if necessary.'
    )
    @click.option(
        '--force_load_normalizers_db',
        '-f',
        is_flag=True,
        default=False,
        help=('Load all normalizers data into database. Overrides '
              '--load_normalizers_db if both are selected.')
    )
    @click.option(
        '--normalizers_db_url',
        default='http://localhost:8000',
        help=('URL endpoint of normalizers DynamoDB database. Set to '
              '`http://localhost:8000` by default.')
    )
    @click.option(
        "--load_latest_cdms",
        "-l",
        is_flag=True,
        default=False,
        help=("Clear MetaKB database and load most recent available source "
              "CDM files. Does not run harvest and transform methods to "
              "generate new CDM files.")
    )
    @click.option(
        "--load_target_cdm",
        "-t",
        type=click.Path(exists=True, dir_okay=False, readable=True,
                        path_type=Path),
        required=False,
        help=("Load transformed CDM file at specified path. Overrides "
              "--load_normalizers_db, --force_load_normalizers_db, "
              "and --load_latest_cdms.")
    )
    def update_metakb_db(db_url: str, db_username: str, db_password: str,
                         load_normalizers_db: bool,
                         force_load_normalizers_db: bool,
                         normalizers_db_url: str,
                         load_latest_cdms: bool,
                         load_target_cdm: Optional[Path]):
        """Execute data harvest and transformation from resources and upload
        to graph datastore.
        """
        db_url = CLI()._check_db_param(db_url, 'URL')
        db_username = CLI()._check_db_param(db_username, 'username')
        db_password = CLI()._check_db_param(db_password, 'password')

        if normalizers_db_url:
            for env_var_name in ['GENE_NORM_DB_URL', 'THERAPY_NORM_DB_URL',
                                 'DISEASE_NORM_DB_URL']:
                environ[env_var_name] = normalizers_db_url

        if not (load_latest_cdms or load_target_cdm):
            if load_normalizers_db or force_load_normalizers_db:
                CLI()._load_normalizers_db(force_load_normalizers_db)

            CLI()._harvest_sources()
            CLI()._transform_sources()

        # Load neo4j database
        start = timer()
        msg = "Loading neo4j database..."
        click.echo(msg)
        logger.info(msg)
        g = Graph(uri=db_url, credentials=(db_username, db_password))
        if load_target_cdm:
            g.load_from_json(load_target_cdm)
        else:
            g.clear()
            for src in sorted({v.value for v
                               in SourceName.__members__.values()}):
                pattern = f"{src}_cdm_*.json"
                globbed = (APP_ROOT / "data" / src / "transform").glob(pattern)
                try:
                    path = sorted(globbed)[-1]
                except IndexError:
                    raise FileNotFoundError(f"No valid transform file found "
                                            f"for {src}")
                g.load_from_json(path)
        g.close()
        end = timer()
        msg = f"Successfully loaded neo4j database in {(end-start):.5f} s"
        click.echo(f"{msg}\n")
        logger.info(msg)

    @staticmethod
    def _harvest_sources() -> None:
        """Run harvesting procedure for all sources."""
        logger.info("Harvesting sources...")
        # TODO: Switch to using constant
        harvester_sources = {
            'civic': CIViCHarvester,
            'moa': MOAHarvester
        }
        total_start = timer()
        for source_str, source_class in harvester_sources.items():
            harvest_start = f"Harvesting {source_str}..."
            click.echo(harvest_start)
            logger.info(harvest_start)
            start = timer()
            source: Harvester = source_class()
            source_successful = source.harvest()
            end = timer()
            if not source_successful:
                logger.info(f'{source_str} harvest failed.')
                click.get_current_context().exit()
            harvest_finish = \
                f"{source_str} harvest finished in {(end - start):.5f} s"
            click.echo(harvest_finish)
            logger.info(harvest_finish)
        total_end = timer()
        msg = f"Successfully harvested all sources in " \
              f"{(total_end - total_start):.5f} s"
        click.echo(f"{msg}\n")
        logger.info(msg)

    @staticmethod
    def _transform_sources() -> None:
        """Run transformation procedure for all sources."""
        logger.info("Transforming harvested data to CDM...")
        # TODO: Switch to using constant
        transform_sources = {
            'civic': CIViCTransform,
            'moa': MOATransform
        }
        total_start = timer()
        for src_str, src_name in transform_sources.items():
            transform_start = f"Transforming {src_str}..."
            click.echo(transform_start)
            logger.info(transform_start)
            start = timer()
            source: Transform = src_name()
            source.transform()
            end = timer()
            transform_end = \
                f"{src_str} transform finished in {(end - start):.5f} s."
            click.echo(transform_end)
            logger.info(transform_end)
            source.create_json()
        total_end = timer()
        msg = f"Successfully transformed all sources to CDM in " \
              f"{(total_end-total_start):.5f} s"
        click.echo(f"{msg}\n")
        logger.info(msg)

    def _load_normalizers_db(self, load_normalizer_db):
        """Load normalizer database source data.

        :param bool load_normalizer_db: Load normalizer database for each
            normalizer
        """
        if load_normalizer_db:
            load_disease = load_therapy = load_gene = True
        else:
            load_disease = self._check_normalizer(
                DiseaseDatabase(), {src.value for src in DiseaseSources})
            load_therapy = self._check_normalizer(
                TherapyDatabase(), {src for src in TherapySources})
            load_gene = self._check_normalizer(
                GeneDatabase(), {src.value for src in GeneSources})

        for load_source, normalizer_cli in [
            (load_disease, DiseaseCLI), (load_therapy, TherapyCLI),
            (load_gene, GeneCLI)
        ]:
            name = \
                str(normalizer_cli).split()[1].split('.')[0][1:].capitalize()
            self._update_normalizer_db(name, load_source, normalizer_cli)
        click.echo("Normalizers database loaded.\n")

    @staticmethod
    def _check_normalizer(db, sources) -> bool:
        """Check whether or not normalizer data needs to be loaded.

        :param Database db: Normalizer database
        :param set sources: Sources that are needed in the normalizer db
        :return: `True` If normalizer needs to be loaded. `False` otherwise.
        """
        for src in sources:
            response = db.metadata.get_item(
                Key={'src_name': src}
            )
            if not response.get('Item'):
                return True
        return False

    @staticmethod
    def _update_normalizer_db(name, load_normalizer, source_cli) -> None:
        """Update Normalizer database.

        :param str name: Name of the normalizer
        :param bool load_normalizer: Whether or not to load normalizer db
        :param CLI source_cli: Normalizer CLI class for loading and
            deleting source data
        """
        if load_normalizer:
            try:
                click.echo(f'\nLoading {name} Normalizer data...')
                source_cli.update_normalizer_db(
                    ['--update_all', '--update_merged'])
                click.echo(f'Successfully Loaded {name} Normalizer data.\n')
            except SystemExit as e:
                if e.code != 0:
                    raise e
        else:
            click.echo(f'{name} Normalizer is already loaded.\n')

    @staticmethod
    def _check_db_param(param: str, name: str) -> str:
        """Check for MetaKB database parameter.
        :param str param: value of parameter as received from command line
        :param str name: name of parameter
        :return: parameter value, or exit with error message if unavailable
        """
        if not param:
            env_var_name = f'METAKB_DB_{name.upper()}'
            if env_var_name in environ.keys():
                return environ[env_var_name]
            else:
                # Default is local
                if name == 'URL':
                    return "bolt://localhost:7687"
                elif name == 'username':
                    return 'neo4j'
                else:
                    return 'admin'
        else:
            return param

    @staticmethod
    def _help_msg(msg: str = ""):
        """Handle invalid user input.
        :param str msg: Error message to display to user.
        """
        ctx = click.get_current_context()
        logger.fatal(f'Error: {msg}')
        click.echo(ctx.get_help())
        ctx.exit()


if __name__ == '__main__':
    CLI().update_metakb_db()



"""A module for the Harvester base class"""
from typing import List, Dict, Optional
import json
import logging
from datetime import datetime as dt

from metakb import APP_ROOT, DATE_FMT

logger = logging.getLogger('metakb')
logger.setLevel(logging.DEBUG)


class Harvester:
    """A base class for content harvesters."""

    def __init__(self):
        """Initialize Harvester class."""
        self.assertions = []

    def harvest(self):
        """
        Retrieve and store records from a resource. Records may be stored in
        any manner, but must be retrievable by :method:`iterate_records`.

        :return: `True` if operation was successful, `False` otherwise.
        :rtype: bool
        """
        raise NotImplementedError

    def iter_assertions(self):
        """
        Yield all :class:`ClinSigAssertion` records for the resource.

        :return: An iterator
        :rtype: Iterator[:class:`ClinSigAssertion`]
        """
        for statement in self.assertions:
            yield statement

    def create_json(self, items: Dict[str, List],
                    filename: Optional[str] = None) -> bool:
        """Create composite and individual JSON for harvested data.

        :param Dict items: item types keyed to Lists of values
        :param Optional[str] filename: custom filename for composite document
        :return: `True` if JSON creation was successful. `False` otherwise.
        """
        composite_dict = dict()
        src = self.__class__.__name__.lower().split("harvest")[0]
        src_dir = APP_ROOT / "data" / src / "harvester"
        src_dir.mkdir(exist_ok=True, parents=True)
        today = dt.strftime(dt.today(), DATE_FMT)
        try:
            for item_type, item_list in items.items():
                composite_dict[item_type] = item_list

                with open(src_dir / f"{item_type}_{today}.json", "w+") as f:
                    f.write(json.dumps(item_list, indent=4))
            if filename is None:
                filename = f"{src}_harvester_{today}.json"
            with open(src_dir / filename, "w+") as f:
                json.dump(composite_dict, f, indent=4)
        except Exception as e:
            logger.error(f"Unable to create json: {e}")
            return False
        return True



"""Common data model"""
from enum import Enum, IntEnum
from typing import List, Optional, Union, Dict, Any, Type

from ga4gh.vrsatile.pydantic.vrs_models import CURIE
from ga4gh.vrsatile.pydantic.vrsatile_models import ValueObjectDescriptor, \
    GeneDescriptor, VariationDescriptor
from pydantic import BaseModel
from pydantic.types import StrictBool

from metakb.version import __version__, LAST_UPDATED


class SourceName(str, Enum):
    """Resources we import directly."""

    CIVIC = 'civic'
    MOA = 'moa'


class XrefSystem(str, Enum):
    """Define constraints for System in xrefs."""

    CLINVAR = 'clinvar'
    CLINGEN = 'caid'
    DB_SNP = 'dbsnp'
    NCBI = 'ncbigene'
    DISEASE_ONTOLOGY = 'do'


class SourcePrefix(str, Enum):
    """Define constraints for source prefixes."""

    PUBMED = 'pmid'
    ASCO = 'asco'


class NormalizerPrefix(str, Enum):
    """Define constraints for normalizer prefixes."""

    GENE = 'gene'


class PropositionType(str, Enum):
    """Define constraints for proposition type."""

    PREDICTIVE = 'therapeutic_response_proposition'
    DIAGNOSTIC = 'diagnostic_proposition'
    PROGNOSTIC = 'prognostic_proposition'
    PREDISPOSING = 'predisposition_proposition'
    FUNCTIONAL = 'functional_consequence_proposition'
    ONCOGENIC = 'oncogenicity_proposition'
    PATHOGENIC = 'pathogenicity_proposition'


class PredictivePredicate(str, Enum):
    """Define constraints for predictive predicate."""

    SENSITIVITY = 'predicts_sensitivity_to'
    RESISTANCE = 'predicts_resistance_to'


class DiagnosticPredicate(str, Enum):
    """Define constraints for diagnostic predicate."""

    POSITIVE = 'is_diagnostic_inclusion_criterion_for'
    NEGATIVE = 'is_diagnostic_exclusion_criterion_for'


class PrognosticPredicate(str, Enum):
    """Define constraints for prognostic predicate."""

    BETTER_OUTCOME = 'is_prognostic_of_better_outcome_for'
    POOR_OUTCOME = 'is_prognostic_of_worse_outcome_for'


class PathogenicPredicate(str, Enum):
    """Define constraints for the pathogenicity predicate."""

    UNCERTAIN_SIGNIFICANCE = 'is_of_uncertain_significance_for'
    PATHOGENIC = 'is_pathogenic_for'
    BENIGN = 'is_benign_for'


class FunctionalPredicate(str, Enum):
    """Define constraints for functional predicate."""

    GAIN_OF_FUNCTION = 'causes_gain_of_function_of'
    LOSS_OF_FUNCTION = 'causes_loss_of_function_of'
    UNALTERED_FUNCTION = 'does_not_change_function_of'
    NEOMORPHIC = 'causes_neomorphic_function_of'
    DOMINATE_NEGATIVE = 'causes_dominant_negative_function_of'


Predicate = Union[PredictivePredicate, DiagnosticPredicate,
                  PrognosticPredicate, PathogenicPredicate,
                  FunctionalPredicate]


class VariationOrigin(str, Enum):
    """Define constraints for variant origin."""

    SOMATIC = 'somatic'
    GERMLINE = 'germline'
    NOT_APPLICABLE = 'N/A'


class Direction(str, Enum):
    """Define constraints for evidence direction."""

    SUPPORTS = 'supports'
    DOES_NOT_SUPPORT = 'does_not_support'


class MoleculeContext(str, Enum):
    """Define constraints for types of molecule context."""

    GENOMIC = 'genomic'
    TRANSCRIPT = 'transcript'
    PROTEIN = 'protein'


class Proposition(BaseModel):
    """Define Proposition model."""

    id: CURIE
    type: PropositionType
    predicate: Predicate
    subject: CURIE  # vrs:Variation
    object_qualifier: CURIE  # vicc:Disease


class TherapeuticResponseProposition(Proposition):
    """Define therapeutic Response Proposition model"""

    type = PropositionType.PREDICTIVE
    predicate: PredictivePredicate
    object: CURIE  # vicc:Therapy

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['TherapeuticResponseProposition']) \
                -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "id": "proposition:133",
                "type": "therapeutic_response_proposition",
                "predicate": "predicts_sensitivity_to",
                "subject": "ga4gh:VA.kgjrhgf84CEndyLjKdAO0RxN-e3pJjxA",
                "object_qualifier": "ncit:C2926",
                "object": "rxcui:1430438"
            }


class PrognosticProposition(Proposition):
    """Defines the Prognostic Proposition model."""

    type = PropositionType.PROGNOSTIC
    predicate: PrognosticPredicate


class DiagnosticProposition(Proposition):
    """Defines the Diagnostic Proposition model."""

    type = PropositionType.DIAGNOSTIC
    predicate: DiagnosticPredicate


class MethodID(IntEnum):
    """Create AssertionMethod id constants for harvested sources."""

    CIVIC_EID_SOP = 1
    CIVIC_AID_AMP_ASCO_CAP = 2
    CIVIC_AID_ACMG = 3
    MOA_ASSERTION_BIORXIV = 4


class Statement(BaseModel):
    """Define Statement model."""

    id: CURIE
    type = 'Statement'
    description: str
    direction: Optional[Direction]
    evidence_level: CURIE
    proposition: CURIE
    variation_origin: Optional[VariationOrigin]
    variation_descriptor: CURIE
    therapy_descriptor: Optional[CURIE]
    disease_descriptor: CURIE
    method: CURIE
    supported_by: List[CURIE]
    # contribution: str  TODO: After metakb first pass


class Document(BaseModel):
    """Define model for Source."""

    id: CURIE
    document_id: Optional[CURIE]
    label: str
    description: Optional[str]
    xrefs: Optional[List[CURIE]]
    type = 'Document'


class Date(BaseModel):
    """Define model for date."""

    year: int
    month: Optional[int]
    day: Optional[int]

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['StatementResponse']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "year": 2019,
                "month": 11,
                "day": 29
            }


class Method(BaseModel):
    """Define model for methods used in evidence curation and classifications."""  # noqa: E501

    id: CURIE
    label: str
    url: str
    version: Date
    authors: str
    type = 'Method'


class Response(BaseModel):
    """Define the Response Model."""

    statements: List[Statement]
    propositions: List[Union[TherapeuticResponseProposition,
                             PrognosticProposition,
                             DiagnosticProposition]]
    variation_descriptors: List[VariationDescriptor]
    gene_descriptors: List[GeneDescriptor]
    therapy_descriptors: Optional[List[ValueObjectDescriptor]]
    disease_descriptors: List[ValueObjectDescriptor]
    methods: List[Method]
    documents: List[Document]


class StatementResponse(BaseModel):
    """Define Statement Response for Search Endpoint."""

    id: CURIE
    type = 'Statement'
    description: str
    direction: Optional[Direction]
    evidence_level: CURIE
    variation_origin: Optional[VariationOrigin]
    proposition: CURIE
    variation_descriptor: CURIE
    therapy_descriptor: Optional[CURIE]
    disease_descriptor: CURIE
    method: CURIE
    supported_by: List[CURIE]

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['StatementResponse']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "id": "civic.eid:2997",
                "description": "Afatinib, an irreversible inhibitor of the ErbB family of tyrosine kinases has been approved in the US for the first-line treatment of patients with metastatic non-small-cell lung cancer (NSCLC) who have tumours with EGFR exon 19 deletions or exon 21 (L858R) substitution mutations as detected by a US FDA-approved test",  # noqa: E501
                "direction": "supports",
                "evidence_level": "civic.evidence_level:A",
                "variation_origin": "somatic",
                "proposition": "proposition:133",
                "variation_descriptor": "civic.vid:33",
                "therapy_descriptor": "civic.tid:146",
                "disease_descriptor": "civic.did:8",
                "method": "method:001",
                "supported_by": [
                    "pmid:23982599"
                ],
                "type": "Statement"
            }


class NestedStatementResponse(BaseModel):
    """Define Statement Response for Search Endpoint."""

    id: CURIE
    type = 'Statement'
    description: str
    direction: Optional[Direction]
    evidence_level: CURIE
    variation_origin: Optional[VariationOrigin]
    proposition: Union[TherapeuticResponseProposition,
                       PrognosticProposition,
                       DiagnosticProposition]
    variation_descriptor: VariationDescriptor
    therapy_descriptor: Optional[ValueObjectDescriptor]
    disease_descriptor: ValueObjectDescriptor
    method: Method
    supported_by: List[Union[Document, CURIE]]

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['NestedStatementResponse']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "id": "civic.eid:2997",
                "description": "Afatinib, an irreversible inhibitor of the ErbB family of tyrosine kinases has been approved in the US for the first-line treatment of patients with metastatic non-small-cell lung cancer (NSCLC) who have tumours with EGFR exon 19 deletions or exon 21 (L858R) substitution mutations as detected by a US FDA-approved test",  # noqa: E501
                "direction": "supports",
                "evidence_level": "civic.evidence_level:A",
                "variation_origin": "somatic",
                "proposition": {
                    "id": "proposition:133",
                    "type": "therapeutic_response_proposition",
                    "predicate": "predicts_sensitivity_to",
                    "subject": "ga4gh:VA.kgjrhgf84CEndyLjKdAO0RxN-e3pJjxA",
                    "object_qualifier": "ncit:C2926",
                    "object": "rxcui:1430438"
                },
                "variation_descriptor": {
                    "id": "civic.vid:33",
                    "type": "VariationDescriptor",
                    "label": "L858R",
                    "description": "EGFR L858R has long been recognized as a functionally significant mutation in cancer, and is one of the most prevalent single mutations in lung cancer. Best described in non-small cell lung cancer (NSCLC), the mutation seems to confer sensitivity to first and second generation TKI's like gefitinib and neratinib. NSCLC patients with this mutation treated with TKI's show increased overall and progression-free survival, as compared to chemotherapy alone. Third generation TKI's are currently in clinical trials that specifically focus on mutant forms of EGFR, a few of which have shown efficacy in treating patients that failed to respond to earlier generation TKI therapies.",  # noqa: E501
                    "xrefs": [
                        "clinvar:376280",
                        "clinvar:16609",
                        "clinvar:376282",
                        "caid:CA126713",
                        "dbsnp:121434568"
                    ],
                    "alternate_labels": [
                        "LEU858ARG"
                    ],
                    "extensions": [
                        {
                            "type": "Extension",
                            "name": "civic_representative_coordinate",
                            "value": {
                                "chromosome": "7",
                                "start": 55259515,
                                "stop": 55259515,
                                "reference_bases": "T",
                                "variant_bases": "G",
                                "representative_transcript":
                                    "ENST00000275493.2",
                                "ensembl_version": 75,
                                "reference_build": "GRCh37"
                            }
                        },
                        {
                            "type": "Extension",
                            "name": "civic_actionability_score",
                            "value": 375
                        }
                    ],
                    "variation_id":
                        "ga4gh:VA.kgjrhgf84CEndyLjKdAO0RxN-e3pJjxA",
                    "variation": {
                        "_id": "ga4gh:VA.kgjrhgf84CEndyLjKdAO0RxN-e3pJjxA",
                        "type": "Allele",
                        "location": {
                            "_id": "ga4gh:VSL.Sfs_3PlVEYp9BxBsHsFfU1tvhfDq361f",  # noqa: E501
                            "type": "SequenceLocation",
                            "sequence_id":
                                "ga4gh:SQ.vyo55F6mA6n2LgN4cagcdRzOuh38V4mE",
                            "interval": {
                                "type": "SequenceInterval",
                                "start": {
                                    "type": "Number",
                                    "value": 857
                                },
                                "end": {
                                    "type": "Number",
                                    "value": 858
                                }
                            }
                        },
                        "state": {
                            "type": "LiteralSequenceExpression",
                            "sequence": "R"
                        }
                    },
                    "structural_type": "SO:0001583",
                    "expressions": [
                        {
                            "type": "Expression",
                            "syntax": "hgvs:genomic",
                            "value": "NC_000007.13:g.55259515T>G"
                        },
                        {
                            "type": "Expression",
                            "syntax": "hgvs:protein",
                            "value": "NP_005219.2:p.Leu858Arg"
                        },
                        {
                            "type": "Expression",
                            "syntax": "hgvs:transcript",
                            "value": "NM_005228.4:c.2573T>G"
                        },
                        {
                            "type": "Expression",
                            "syntax": "hgvs:transcript",
                            "value": "ENST00000275493.2:c.2573T>G"
                        }
                    ],
                    "gene_context": {
                        "id": "civic.gid:19",
                        "type": "GeneDescriptor",
                        "label": "EGFR",
                        "description": "EGFR is widely recognized for its importance in cancer. Amplification and mutations have been shown to be driving events in many cancer types. Its role in non-small cell lung cancer, glioblastoma and basal-like breast cancers has spurred many research and drug development efforts. Tyrosine kinase inhibitors have shown efficacy in EGFR amplfied tumors, most notably gefitinib and erlotinib. Mutations in EGFR have been shown to confer resistance to these drugs, particularly the variant T790M, which has been functionally characterized as a resistance marker for both of these drugs. The later generation TKI's have seen some success in treating these resistant cases, and targeted sequencing of the EGFR locus has become a common practice in treatment of non-small cell lung cancer. \nOverproduction of ligands is another possible mechanism of activation of EGFR. ERBB ligands include EGF, TGF-a, AREG, EPG, BTC, HB-EGF, EPR and NRG1-4 (for detailed information please refer to the respective ligand section).",  # noqa: E501
                        "xrefs": [
                            "ncbigene:1956"
                        ],
                        "alternate_labels": [
                            "ERRP",
                            "EGFR",
                            "mENA",
                            "PIG61",
                            "NISBD2",
                            "HER1",
                            "ERBB1",
                            "ERBB"
                        ],
                        "gene_id": "hgnc:3236"
                    }
                },
                "therapy_descriptor": {
                    "id": "civic.tid:146",
                    "type": "TherapyDescriptor",
                    "label": "Afatinib",
                    "xrefs": [
                        "ncit:C66940"
                    ],
                    "alternate_labels": [
                        "BIBW2992",
                        "BIBW 2992",
                        "(2e)-N-(4-(3-Chloro-4-Fluoroanilino)-7-(((3s)-Oxolan-3-yl)Oxy)Quinoxazolin-6-yl)-4-(Dimethylamino)But-2-Enamide"  # noqa: E501
                    ],
                    "therapy_id": "rxcui:1430438"
                },
                "disease_descriptor": {
                    "id": "civic.did:8",
                    "type": "DiseaseDescriptor",
                    "label": "Lung Non-small Cell Carcinoma",
                    "xrefs": [
                        "DOID:3908"
                    ],
                    "disease_id": "ncit:C2926"
                },
                "method": {
                    "id": "method:001",
                    "label": "Standard operating procedure for curation and clinical interpretation of variants in cancer",  # noqa: E501
                    "url": "https://genomemedicine.biomedcentral.com/articles/10.1186/s13073-019-0687-x",  # noqa: E501
                    "version": {
                        "year": 2019,
                        "month": 11,
                        "day": 29
                    },
                    "authors":
                        "Danos, A.M., Krysiak, K., Barnell, E.K. et al.",
                    "type": "Method"
                },
                "supported_by": [
                    {
                        "id": "pmid:23982599",
                        "label": "Dungo et al., 2013, Drugs",
                        "description": "Afatinib: first global approval.",
                        "type": "Document"
                    }
                ],
                "type": "Statement"
            }


class SearchQuery(BaseModel):
    """Queries for the Search Endpoint."""

    variation: Optional[str]
    disease: Optional[str]
    therapy: Optional[str]
    gene: Optional[str]
    statement_id: Optional[str]
    detail: StrictBool

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['SearchQuery']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "variation": "NP_005219.2:p.Leu858Arg",
                "disease": "Lung Non-small Cell Carcinoma",
                "therapy": "Afatinib",
                "statement_id": "civic.eid:2997",
                "detail": False
            }


class SearchStatementsQuery(BaseModel):
    """Queries for the Search Endpoint."""

    variation: Optional[str]
    disease: Optional[str]
    therapy: Optional[str]
    gene: Optional[str]
    statement_id: Optional[str]

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['SearchStatementsQuery']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "variation": "NP_005219.2:p.Leu858Arg",
                "disease": "Lung Non-small Cell Carcinoma",
                "therapy": "Afatinib",
                "statement_id": "civic.eid:2997"
            }


class Matches(BaseModel):
    """Statements and Propositions that match the queried parameters."""

    statements: List[str]
    propositions: List[str]

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['Matches']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "statements": ["civic.eid:2997"],
                "propositions": ["proposition:133"]
            }


class ServiceMeta(BaseModel):
    """Metadata for MetaKB service."""

    name = "metakb"
    version = __version__
    last_updated = LAST_UPDATED
    url = "https://github.com/cancervariants/metakb"

    class Config:
        """Configure schema example."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["ServiceMeta"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "name": "metakb",
                "version": "1.1.0-alpha.4",
                "last_updated": "2021-12-16",
                "url": "https://github.com/cancervariants/metakb"
            }


class SearchService(BaseModel):
    """Define model for Search Endpoint Response."""

    query: SearchQuery
    warnings: Optional[List[str]]
    matches: Matches
    statements: Optional[List[StatementResponse]]
    propositions: Optional[List[Union[TherapeuticResponseProposition,
                                      DiagnosticProposition,
                                      PrognosticProposition]]]
    variation_descriptors: Optional[List[VariationDescriptor]]
    gene_descriptors: Optional[List[GeneDescriptor]]
    therapy_descriptors: Optional[List[ValueObjectDescriptor]]
    disease_descriptors: Optional[List[ValueObjectDescriptor]]
    methods: Optional[List[Method]]
    documents: Optional[List[Document]]
    service_meta_: ServiceMeta

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['SearchService']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "query": {
                    "variation": "EGFR L858R",
                    "disease": "Lung Non-small Cell Carcinoma",
                    "therapy": "Afatinib",
                    "statement_id": "civic.eid:2997",
                    "detail": False
                },
                "warnings": [],
                "matches": {
                    "statements": ["civic.eid:2997"],
                    "propositions": ["proposition:109"]
                },
                "statements": [
                    {
                        "id": "civic.eid:2997",
                        "description": "Afatinib, an irreversible inhibitor of the ErbB family of tyrosine kinases has been approved in the US for the first-line treatment of patients with metastatic non-small-cell lung cancer (NSCLC) who have tumours with EGFR exon 19 deletions or exon 21 (L858R) substitution mutations as detected by a US FDA-approved test",  # noqa: E501
                        "direction": "supports",
                        "evidence_level": "civic.evidence_level:A",
                        "variation_origin": "somatic",
                        "proposition": "proposition:133",
                        "variation_descriptor": "civic.vid:33",
                        "therapy_descriptor": "civic.tid:146",
                        "disease_descriptor": "civic.did:8",
                        "method": "method:001",
                        "supported_by": [
                            "pmid:23982599"
                        ],
                        "type": "Statement"
                    }
                ],
                "propositions": [
                    {
                        "id": "proposition:133",
                        "type": "therapeutic_response_proposition",
                        "predicate": "predicts_sensitivity_to",
                        "subject": "ga4gh:VA.kgjrhgf84CEndyLjKdAO0RxN-e3pJjxA",
                        "object_qualifier": "ncit:C2926",
                        "object": "rxcui:1430438"
                    }
                ],
                "service_meta_": {
                    "name": "metakb",
                    "version": "1.1.0-alpha.4",
                    "last_updated": "2021-12-16",
                    "url": "https://github.com/cancervariants/metakb"
                }
            }


class SearchIDService(BaseModel):
    """Define model for Search by ID Endpoint Response."""

    query: str
    warnings: Optional[List[str]]
    statement: Optional[StatementResponse]
    proposition: Optional[Union[TherapeuticResponseProposition,
                                DiagnosticProposition,
                                PrognosticProposition]]
    variation_descriptor: Optional[VariationDescriptor]
    gene_descriptor: Optional[GeneDescriptor]
    therapy_descriptor: Optional[ValueObjectDescriptor]
    disease_descriptor: Optional[ValueObjectDescriptor]
    document: Optional[Document]
    method: Optional[Method]
    service_meta_: ServiceMeta

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['SearchIDService']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "query": {
                    "node_id": "civic.vid:33"
                },
                "warnings": [],
                "matches": {
                    "node": "civic.vid:33"
                },
                "variation_descriptors": [
                    {
                        "id": "civic.vid:33",
                        "type": "VariationDescriptor",
                        "label": "L858R",
                        "description": "EGFR L858R has long been recognized "
                                       "as a functionally significant "
                                       "mutation in cancer, and is one of "
                                       "he most prevalent single mutations in"
                                       " lung cancer. Best described in "
                                       "non-small cell lung cancer (NSCLC), "
                                       "the mutation seems to confer "
                                       "sensitivity to first and second "
                                       "generation TKI's like gefitinib and"
                                       " neratinib. NSCLC patients with this"
                                       " mutation treated with TKI's show "
                                       "increased overall and "
                                       "progression-free survival, as "
                                       "compared to chemotherapy alone. "
                                       "Third generation TKI's are currently"
                                       " in clinical trials that specifically"
                                       " focus on mutant forms of EGFR, a few"
                                       " of which have shown efficacy in "
                                       "treating patients that failed to "
                                       "respond to earlier generation "
                                       "TKI therapies.",
                        "value_id": "ga4gh:VA.WyOqFMhc8aOnMFgdY0uM7nSLNqxVPAiR",  # noqa: E501
                        "value": {
                            "location": {
                                "interval": {
                                    "end": 858,
                                    "start": 857,
                                    "type": "SimpleInterval"
                                },
                                "sequence_id": "ga4gh:SQ.vyo55F6mA6n2LgN4cagcdRzOuh38V4mE",  # noqa: E501
                                "type": "SequenceLocation"
                            },
                            "state": {
                                "sequence": "R",
                                "type": "SequenceState"
                            },
                            "type": "Allele"
                        },
                        "xrefs": [
                            "clinvar:376280",
                            "clinvar:16609",
                            "clinvar:376282",
                            "caid:CA126713",
                            "dbsnp:121434568"
                        ],
                        "alternate_labels": [
                            "LEU858ARG"
                        ],
                        "extensions": [
                            {
                                "name": "civic_representative_coordinate",
                                "value": {
                                    "chromosome": "7",
                                    "start": 55259515,
                                    "stop": 55259515,
                                    "reference_bases": "T",
                                    "variant_bases": "G",
                                    "representative_transcript": "ENST00000275493.2",  # noqa: E501
                                    "ensembl_version": 75,
                                    "reference_build": "GRCh37"
                                },
                                "type": "Extension"
                            },
                            {
                                "name": "civic_actionability_score",
                                "value": "352.5",
                                "type": "Extension"
                            }
                        ],
                        "structural_type": "SO:0001583",
                        "expressions": [
                            {
                                "syntax": "hgvs:genomic",
                                "value": "NC_000007.13:g.55259515T>G",
                                "type": "Expression"
                            },
                            {
                                "syntax": "hgvs:protein",
                                "value": "NP_005219.2:p.Leu858Arg",
                                "type": "Expression"
                            },
                            {
                                "syntax": "hgvs:transcript",
                                "value": "NM_005228.4:c.2573T>G",
                                "type": "Expression"
                            },
                            {
                                "syntax": "hgvs:transcript",
                                "value": "ENST00000275493.2:c.2573T>G",
                                "type": "Expression"
                            }
                        ],
                        "gene_context": "civic.gid:19"
                    }
                ],
                "service_meta_": {
                    "name": "metakb",
                    "version": "1.1.0-alpha.4",
                    "last_updated": "2021-12-16",
                    "url": "https://github.com/cancervariants/metakb"
                }
            }


class SearchStatementsService(BaseModel):
    """Define model for Search Statements Endpoint Response."""

    query: SearchStatementsQuery
    warnings: Optional[List[str]]
    matches: Matches
    statements: Optional[List[NestedStatementResponse]]
    service_meta_: ServiceMeta

    class Config:
        """Configure examples."""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type['SearchStatementsService']) -> None:
            """Configure OpenAPI schema"""
            if 'title' in schema.keys():
                schema.pop('title', None)
            for prop in schema.get('properties', {}).values():
                prop.pop('title', None)
            schema['example'] = {
                "query": {
                    "variation": "EGFR L858R",
                    "disease": "Lung Non-small Cell Carcinoma",
                    "therapy": "Afatinib",
                    "statement_id": "civic.eid:2997"
                },
                "warnings": [],
                "matches": {
                    "statements": ["civic.eid:2997"],
                    "propositions": ["proposition:109"]
                },
                "statements": [
                    {
                        "id": "civic.eid:2997",
                        "description": "Afatinib, an irreversible inhibitor of the ErbB family of tyrosine kinases has been approved in the US for the first-line treatment of patients with metastatic non-small-cell lung cancer (NSCLC) who have tumours with EGFR exon 19 deletions or exon 21 (L858R) substitution mutations as detected by a US FDA-approved test",  # noqa: E501
                        "direction": "supports",
                        "evidence_level": "civic.evidence_level:A",
                        "variation_origin": "somatic",
                        "proposition": {
                            "id": "proposition:133",
                            "type": "therapeutic_response_proposition",
                            "predicate": "predicts_sensitivity_to",
                            "subject": "ga4gh:VA.kgjrhgf84CEndyLjKdAO0RxN-e3pJjxA",    # noqa: E501
                            "object_qualifier": "ncit:C2926",
                            "object": "rxcui:1430438"
                        },
                        "variation_descriptor": {
                            "id": "civic.vid:33",
                            "type": "VariationDescriptor",
                            "label": "L858R",
                            "description": "EGFR L858R has long been recognized as a functionally significant mutation in cancer, and is one of the most prevalent single mutations in lung cancer. Best described in non-small cell lung cancer (NSCLC), the mutation seems to confer sensitivity to first and second generation TKI's like gefitinib and neratinib. NSCLC patients with this mutation treated with TKI's show increased overall and progression-free survival, as compared to chemotherapy alone. Third generation TKI's are currently in clinical trials that specifically focus on mutant forms of EGFR, a few of which have shown efficacy in treating patients that failed to respond to earlier generation TKI therapies.",  # noqa: E501
                            "xrefs": [
                                "clinvar:376280",
                                "clinvar:16609",
                                "clinvar:376282",
                                "caid:CA126713",
                                "dbsnp:121434568"
                            ],
                            "alternate_labels": [
                                "LEU858ARG"
                            ],
                            "extensions": [
                                {
                                    "type": "Extension",
                                    "name": "civic_representative_coordinate",
                                    "value": {
                                        "chromosome": "7",
                                        "start": 55259515,
                                        "stop": 55259515,
                                        "reference_bases": "T",
                                        "variant_bases": "G",
                                        "representative_transcript":
                                            "ENST00000275493.2",
                                        "ensembl_version": 75,
                                        "reference_build": "GRCh37"
                                    }
                                },
                                {
                                    "type": "Extension",
                                    "name": "civic_actionability_score",
                                    "value": 375
                                }
                            ],
                            "variation_id":
                                "ga4gh:VA.kgjrhgf84CEndyLjKdAO0RxN-e3pJjxA",
                            "variation": {
                                "_id": "ga4gh:VA.kgjrhgf84CEndyLjKdAO0RxN-e3pJjxA",  # noqa: E501
                                "type": "Allele",
                                "location": {
                                    "_id": "ga4gh:VSL.Sfs_3PlVEYp9BxBsHsFfU1tvhfDq361f",  # noqa: E501
                                    "type": "SequenceLocation",
                                    "sequence_id": "ga4gh:SQ.vyo55F6mA6n2LgN4cagcdRzOuh38V4mE",  # noqa: E501
                                    "interval": {
                                        "type": "SequenceInterval",
                                        "start": {
                                            "type": "Number",
                                            "value": 857
                                        },
                                        "end": {
                                            "type": "Number",
                                            "value": 858
                                        }
                                    }
                                },
                                "state": {
                                    "type": "LiteralSequenceExpression",
                                    "sequence": "R"
                                }
                            },
                            "structural_type": "SO:0001583",
                            "expressions": [
                                {
                                    "type": "Expression",
                                    "syntax": "hgvs:genomic",
                                    "value": "NC_000007.13:g.55259515T>G"
                                },
                                {
                                    "type": "Expression",
                                    "syntax": "hgvs:protein",
                                    "value": "NP_005219.2:p.Leu858Arg"
                                },
                                {
                                    "type": "Expression",
                                    "syntax": "hgvs:transcript",
                                    "value": "NM_005228.4:c.2573T>G"
                                },
                                {
                                    "type": "Expression",
                                    "syntax": "hgvs:transcript",
                                    "value": "ENST00000275493.2:c.2573T>G"
                                }
                            ],
                            "gene_context": {
                                "id": "civic.gid:19",
                                "type": "GeneDescriptor",
                                "label": "EGFR",
                                "description": "EGFR is widely recognized for its importance in cancer. Amplification and mutations have been shown to be driving events in many cancer types. Its role in non-small cell lung cancer, glioblastoma and basal-like breast cancers has spurred many research and drug development efforts. Tyrosine kinase inhibitors have shown efficacy in EGFR amplfied tumors, most notably gefitinib and erlotinib. Mutations in EGFR have been shown to confer resistance to these drugs, particularly the variant T790M, which has been functionally characterized as a resistance marker for both of these drugs. The later generation TKI's have seen some success in treating these resistant cases, and targeted sequencing of the EGFR locus has become a common practice in treatment of non-small cell lung cancer. \nOverproduction of ligands is another possible mechanism of activation of EGFR. ERBB ligands include EGF, TGF-a, AREG, EPG, BTC, HB-EGF, EPR and NRG1-4 (for detailed information please refer to the respective ligand section).",  # noqa: E501
                                "xrefs": [
                                    "ncbigene:1956"
                                ],
                                "alternate_labels": [
                                    "ERRP",
                                    "EGFR",
                                    "mENA",
                                    "PIG61",
                                    "NISBD2",
                                    "HER1",
                                    "ERBB1",
                                    "ERBB"
                                ],
                                "gene_id": "hgnc:3236"
                            }
                        },
                        "therapy_descriptor": {
                            "id": "civic.tid:146",
                            "type": "TherapyDescriptor",
                            "label": "Afatinib",
                            "xrefs": [
                                "ncit:C66940"
                            ],
                            "alternate_labels": [
                                "BIBW2992",
                                "BIBW 2992",
                                "(2e)-N-(4-(3-Chloro-4-Fluoroanilino)-7-(((3s)-Oxolan-3-yl)Oxy)Quinoxazolin-6-yl)-4-(Dimethylamino)But-2-Enamide"  # noqa: E501
                            ],
                            "therapy_id": "rxcui:1430438"
                        },
                        "disease_descriptor": {
                            "id": "civic.did:8",
                            "type": "DiseaseDescriptor",
                            "label": "Lung Non-small Cell Carcinoma",
                            "xrefs": [
                                "DOID:3908"
                            ],
                            "disease_id": "ncit:C2926"
                        },
                        "method": {
                            "id": "method:001",
                            "label": "Standard operating procedure for curation and clinical interpretation of variants in cancer",  # noqa: E501
                            "url": "https://genomemedicine.biomedcentral.com/articles/10.1186/s13073-019-0687-x",  # noqa: E501
                            "version": {
                                "year": 2019,
                                "month": 11,
                                "day": 29
                            },
                            "authors": "Danos, A.M., Krysiak, K., Barnell, E.K. et al.",  # noqa: E501
                            "type": "Method"
                        },
                        "supported_by": [
                            {
                                "id": "pmid:23982599",
                                "label": "Dungo et al., 2013, Drugs",
                                "description": "Afatinib: first global approval.",  # noqa: E501
                                "type": "Document"
                            }
                        ],
                        "type": "Statement"
                    }
                ],
                "service_meta_": {
                    "name": "metakb",
                    "version": "1.1.0-alpha.4",
                    "last_updated": "2021-12-16",
                    "url": "https://github.com/cancervariants/metakb"
                }
            }



"""A module for to transform CIViC."""
from typing import Optional, Dict, List, Set
from pathlib import Path
import logging

from ga4gh.vrsatile.pydantic.vrsatile_models import VariationDescriptor, \
    Extension, Expression, GeneDescriptor, ValueObjectDescriptor

from metakb import APP_ROOT
from metakb.normalizers import VICCNormalizers
from metakb.transform.base import Transform
import metakb.schemas as schemas


logger = logging.getLogger('metakb.transform.civic')
logger.setLevel(logging.DEBUG)


class CIViCTransform(Transform):
    """A class for transforming CIViC to the common data model."""

    def __init__(self,
                 data_dir: Path = APP_ROOT / "data",
                 harvester_path: Optional[Path] = None,
                 normalizers: Optional[VICCNormalizers] = None) -> None:
        """Initialize CIViC Transform class.
        :param Path data_dir: Path to source data directory
        :param Optional[Path] harvester_path: Path to previously harvested data
        :param VICCNormalizers normalizers: normalizer collection instance
        """
        super().__init__(data_dir=data_dir,
                         harvester_path=harvester_path,
                         normalizers=normalizers)
        # Able to normalize these IDSs
        self.valid_ids = {
            'variation_descriptors': dict(),
            'disease_descriptors': dict(),
            'therapy_descriptors': dict()
        }
        # Unable to normalize these IDs
        self.invalid_ids = {
            'therapy_descriptors': list(),
            'disease_descriptors': list()
        }

    def transform(self):
        """Transform CIViC harvested json to common data model."""
        data = self.extract_harvester()
        evidence_items = data['evidence']
        assertions = data['assertions']
        variants = data['variants']
        genes = data['genes']

        # Filter Variant IDs for
        # Prognostic, Predictive, and Diagnostic evidence
        supported_evidence_types = ['Prognostic', 'Predictive', 'Diagnostic']
        vids = {e['variant_id'] for e in evidence_items
                if e['evidence_type'] in supported_evidence_types}
        vids |= {a['variant']['id'] for a in assertions
                 if a['evidence_type'] in supported_evidence_types}

        self._add_variation_descriptors(variants, vids)
        self._add_gene_descriptors(genes)
        self._add_methods()
        self._transform_evidence_and_assertions(evidence_items)
        self._transform_evidence_and_assertions(assertions, is_evidence=False)

    def _transform_evidence_and_assertions(self, records: List[Dict],
                                           is_evidence=True) -> None:
        """Transform statements, propositions, descriptors, and documents
        from CIViC evidence items and assertions.

        :param list records: CIViC Evidence Items or Assertions
        :param bool is_evidence: `True` if records are evidence items.
            `False` if records are assertions.
        """
        for r in records:
            name_lower = r['name'].lower()
            if name_lower.startswith('eid'):
                civic_id = name_lower.replace('eid', 'civic.eid:')
            else:
                civic_id = name_lower.replace('aid', 'civic.aid:')

            # Omit entries that are not in an accepted state
            if r['status'] != 'accepted':
                logger.warning(f"{civic_id} has status: {r['status']}")
                continue

            if r['evidence_type'] not in ['Predictive', 'Prognostic',
                                          'Diagnostic']:
                continue
            else:
                # Functional Evidence types do not have a disease
                if not r['disease']:
                    continue

            if r['evidence_type'] == 'Predictive':
                if len(r['drugs']) != 1:
                    continue
                else:
                    therapy_id = f"civic.tid:{r['drugs'][0]['id']}"
                    therapy_descriptor = \
                        self._add_therapy_descriptor(therapy_id, r)
                    if not therapy_descriptor:
                        continue

                    if therapy_descriptor not in self.therapy_descriptors:
                        self.therapy_descriptors.append(therapy_descriptor)
            else:
                therapy_id = None
                therapy_descriptor = None

            disease_id = f"civic.did:{r['disease']['id']}"
            disease_descriptor = self._add_disease_descriptor(disease_id, r)
            if not disease_descriptor:
                continue

            if disease_descriptor not in self.disease_descriptors:
                self.disease_descriptors.append(disease_descriptor)

            if is_evidence:
                variant_id = f"civic.vid:{r['variant_id']}"
            else:
                variant_id = f"civic.vid:{r['variant']['id']}"
            variation_descriptor = \
                self.valid_ids['variation_descriptors'].get(variant_id)
            if not variation_descriptor:
                continue

            proposition = self._get_proposition(
                r, variation_descriptor, disease_descriptor,
                therapy_descriptor
            )

            # Only support Therapeutic Response and Prognostic
            if not proposition:
                continue

            if proposition not in self.propositions:
                self.propositions.append(proposition)

            if is_evidence:
                # Evidence items's method and evidence level
                method = f'method:{schemas.MethodID.CIVIC_EID_SOP}'
                evidence_level = f"civic.evidence_level:{r['evidence_level']}"

                # Supported by evidence for evidence item
                document = self._get_eid_document(r['source'])
                if document not in self.documents:
                    self.documents.append(document)
                supported_by = [document['id']]
            else:
                # Assertion's method
                if r['amp_level'] and not r['acmg_codes']:
                    method = \
                        f'method:' \
                        f'{schemas.MethodID.CIVIC_AID_AMP_ASCO_CAP}'
                elif not r['amp_level'] and r['acmg_codes']:
                    method = f'method:' \
                             f'{schemas.MethodID.CIVIC_AID_ACMG}'
                else:
                    # Statements are required to have a method
                    logger.warning(f"Unable to get method for {civic_id}")
                    continue

                # assertion's evidence level
                evidence_level = self._get_assertion_evidence_level(r)

                # Supported by evidence for assertion
                supported_by = list()
                documents = \
                    self._get_aid_document(r)
                for d in documents:
                    if d not in self.documents:
                        self.documents.append(d)
                    supported_by.append(d['id'])
                for evidence_item in r['evidence_items']:
                    supported_by.append(f"civic.eid:"
                                        f"{evidence_item['id']}")

            statement = schemas.Statement(
                id=civic_id,
                description=r['description'],
                direction=self._get_evidence_direction(
                    r['evidence_direction']),
                evidence_level=evidence_level,
                proposition=proposition['id'],
                variation_origin=self._get_variation_origin(
                    r['variant_origin']),
                variation_descriptor=variant_id,
                therapy_descriptor=therapy_id,
                disease_descriptor=disease_id,
                method=method,
                supported_by=supported_by
            ).dict(exclude_none=True)
            self.statements.append(statement)

    def _get_evidence_direction(self, direction) -> Optional[str]:
        """Return the evidence direction.

        :param str direction: The civic evidence_direction value
        :return: `supports` or `does_not_support` or None
        """
        if direction == 'Supports':
            return schemas.Direction.SUPPORTS.value
        elif direction == 'Does Not Support':
            return schemas.Direction.DOES_NOT_SUPPORT.value
        else:
            return None

    def _get_assertion_evidence_level(self, assertion) -> Optional[str]:
        """Return evidence_level for CIViC assertion.

        :param dict assertion: CIViC Assertion
        :return: CIViC assertion evidence_level
        """
        evidence_level = None
        # TODO: CHECK
        if assertion['amp_level']:
            if assertion['amp_level'] == 'Not Applicable':
                evidence_level = None
            else:
                tier, level = assertion['amp_level'].split(' - ')
                tier = tier.split()[1]
                if tier == 'I':
                    tier = 1
                elif tier == 'II':
                    tier = 2
                elif tier == 'III':
                    tier = 3
                elif tier == 'IV':
                    tier = 4
                evidence_level = f"amp_asco_cap_2017_level:" \
                                 f"{tier}{level.split()[1]}"
        return evidence_level

    def _get_proposition(self, record, variation_descriptor,
                         disease_descriptor, therapy_descriptor
                         ) -> Optional[dict]:
        """Return a proposition for a record.

        :param dict record: CIViC EID or AID
        :param dict variation_descriptor: The record's variation descriptor
        :param dict disease_descriptor: The record's disease descriptor
        :param dict therapy_descriptor: The record's therapy descriptor
        :return: A proposition
        """
        try:
            proposition_type = \
                self._get_proposition_type(record["evidence_type"])
        except KeyError:
            return None

        predicate = self._get_predicate(proposition_type,
                                        record["clinical_significance"])

        # Don't support TR that has  `None`, "N/A", or "Unknown" predicate
        if not predicate:
            return None

        params = {
            "id": "",
            "type": proposition_type,
            "predicate": predicate,
            "subject": variation_descriptor["variation_id"],
            "object_qualifier": disease_descriptor["disease_id"]
        }

        if proposition_type == schemas.PropositionType.PREDICTIVE:
            params["object"] = therapy_descriptor["therapy_id"]
            proposition_id = self._get_proposition_id(
                params["type"],
                params["predicate"],
                [params["subject"]],
                [params["object_qualifier"]],
                [params["object"]]
            )
        else:
            proposition_id = self._get_proposition_id(
                params["type"],
                params["predicate"],
                [params["subject"]],
                [params["object_qualifier"]]
            )
        if proposition_id is None:
            return None
        else:
            params["id"] = proposition_id

        if proposition_type == schemas.PropositionType.PROGNOSTIC.value:
            proposition = \
                schemas.PrognosticProposition(**params).dict(exclude_none=True)
        elif proposition_type == schemas.PropositionType.PREDICTIVE.value:
            params["object"] = therapy_descriptor["therapy_id"]
            proposition =\
                schemas.TherapeuticResponseProposition(**params).dict(
                    exclude_none=True
                )
        elif proposition_type == schemas.PropositionType.DIAGNOSTIC.value:
            proposition = \
                schemas.DiagnosticProposition(**params).dict(
                    exclude_none=True)
        else:
            proposition = None
        return proposition

    def _get_proposition_type(self,
                              evidence_type,
                              is_evidence=True) -> Optional[schemas.PropositionType]:  # noqa: E501
        """Return proposition type for a given EID or AID.

        :param str evidence_type: CIViC evidence type
        :param bool is_evidence: `True` if EID. `False` if AID.
        :return: Proposition's type
        """
        evidence_type = evidence_type.upper()
        if evidence_type in schemas.PropositionType.__members__.keys():
            if evidence_type == 'PREDISPOSING':
                if is_evidence:
                    proposition_type = schemas.PropositionType.PREDISPOSING
                else:
                    proposition_type = schemas.PropositionType.PATHOGENIC
            else:
                proposition_type = schemas.PropositionType[evidence_type]
        else:
            raise KeyError(f"Proposition Type {evidence_type} not found in "
                           f"schemas.PropositionType")
        return proposition_type

    def _get_variation_origin(self, variant_origin) -> Optional[str]:
        """Return variant origin.

        :param str variant_origin: CIViC variant origin
        :return: Variation origin
        """
        if variant_origin == 'Somatic':
            origin = schemas.VariationOrigin.SOMATIC.value
        elif variant_origin in ['Rare Germline', 'Common Germline']:
            origin = schemas.VariationOrigin.GERMLINE.value
        elif variant_origin == 'N/A':
            origin = schemas.VariationOrigin.NOT_APPLICABLE.value
        else:
            origin = None
        return origin

    def _get_predicate(self, proposition_type,
                       clin_sig) -> Optional[schemas.Predicate]:
        """Return predicate for an evidence item.

        :param str proposition_type: The proposition type
        :param str clin_sig: The evidence item's clinical significance
        :return: Predicate for proposition if valid
        """
        if clin_sig is None or clin_sig.upper() in ['N/A', 'UNKNOWN']:
            return None

        clin_sig = '_'.join(clin_sig.upper().split())
        predicate = None

        if proposition_type == schemas.PropositionType.PREDICTIVE:
            if clin_sig == 'SENSITIVITY/RESPONSE':
                predicate = schemas.PredictivePredicate.SENSITIVITY
            elif clin_sig == 'RESISTANCE':
                predicate = schemas.PredictivePredicate.RESISTANCE
        elif proposition_type == schemas.PropositionType.DIAGNOSTIC:
            predicate = schemas.DiagnosticPredicate[clin_sig]
        elif proposition_type == schemas.PropositionType.PROGNOSTIC:
            if clin_sig == 'POSITIVE':
                predicate = schemas.PrognosticPredicate.BETTER_OUTCOME
            else:
                predicate = schemas.PrognosticPredicate[clin_sig]
        elif proposition_type == schemas.PropositionType.FUNCTIONAL:
            predicate = schemas.FunctionalPredicate[clin_sig]
        elif proposition_type == schemas.PropositionType.ONCOGENIC:
            # TODO: There are currently no Oncogenic types in CIViC harvester
            #  Look into why this is
            pass
        elif proposition_type == schemas.PropositionType.PATHOGENIC:
            if clin_sig in ['PATHOGENIC', 'LIKELY_PATHOGENIC']:
                predicate = schemas.PathogenicPredicate.PATHOGENIC
        else:
            logger.warning(f"CIViC proposition type: {proposition_type} "
                           f"not supported in Predicate schemas")
        return predicate

    def _add_variation_descriptors(self, variants: List, vids: Set) -> None:
        """Add Variation Descriptors to dict of transformations.

        :param List variants: CIViC variants
        :param set vids: Candidate CIViC Variant IDs
        """
        for variant in variants:
            if variant["id"] not in vids:
                continue
            variant_id = f"civic.vid:{variant['id']}"
            if "c." in variant["name"]:
                variant_name = variant["name"]
                if "(" in variant_name:
                    variant_name = \
                        variant_name.replace("(", "").replace(")", "")
                variant_name = variant_name.split()[-1]
            else:
                variant_name = variant["name"]

            variant_query = f"{variant['entrez_name']} {variant_name}"
            hgvs_exprs = self._get_hgvs_expr(variant)

            # TODO: Remove as more get implemented in variation normalizer
            #  Filtering to speed up transformation
            vname_lower = variant["name"].lower()

            if vname_lower.endswith("fs") or "-" in vname_lower or "/" in vname_lower:  # noqa: E501
                if not hgvs_exprs:
                    logger.warning("Variation Normalizer does not support "
                                   f"{variant_id}: {variant_query}")
                    continue

            unable_to_normalize = {
                "mutation", "amplification", "exon", "overexpression",
                "frameshift", "promoter", "deletion", "type", "insertion",
                "expression", "duplication", "copy", "underexpression",
                "number", "variation", "repeat", "rearrangement", "activation",
                "expression", "mislocalization", "translocation", "wild",
                "polymorphism", "frame", "shift", "loss", "function", "levels",
                "inactivation", "snp", "fusion", "dup", "truncation",
                "homozygosity", "gain", "phosphorylation",
            }

            if set(vname_lower.split()) & unable_to_normalize:
                logger.warning("Variation Normalizer does not support "
                               f"{variant_id}: {variant_query}")
                continue

            variation_norm_resp = self.vicc_normalizers.normalize_variation(
                [variant_query]
            )

            # Couldn't find normalized concept
            if not variation_norm_resp:
                logger.warning("Variation Normalizer unable to normalize "
                               f"civic.vid:{variant['id']} using query "
                               f"{variant_query}")
                continue

            if variant["variant_types"]:
                structural_type = variant["variant_types"][0]["so_id"]
            else:
                structural_type = None

            variation_descriptor = VariationDescriptor(
                id=variant_id,
                label=variant["name"],
                description=variant["description"] if variant["description"] else None,  # noqa: E501
                variation_id=variation_norm_resp.variation_id,
                variation=variation_norm_resp.variation,
                gene_context=f"civic.gid:{variant['gene_id']}",
                structural_type=structural_type,
                expressions=hgvs_exprs,
                xrefs=self._get_variant_xrefs(variant),
                alternate_labels=[v_alias for v_alias in
                                  variant["variant_aliases"] if not
                                  v_alias.startswith("RS")],
                extensions=self._get_variant_extensions(variant)
            ).dict(by_alias=True, exclude_none=True)
            self.valid_ids["variation_descriptors"][variant_id] = \
                variation_descriptor
            self.variation_descriptors.append(
                variation_descriptor
            )

    def _get_variant_extensions(self, variant) -> list:
        """Return a list of extensions for a variant.

        :param dict variant: A CIViC variant record
        :return: A list of extensions
        """
        extensions = [
            Extension(
                name='civic_representative_coordinate',
                value={k: v for k, v in variant['coordinates'].items()
                       if v is not None}
            ).dict(exclude_none=True),
            Extension(
                name='civic_actionability_score',
                value=variant['civic_actionability_score']
            ).dict(exclude_none=True)
        ]

        variant_groups = variant['variant_groups']
        if variant_groups:
            v_groups = list()
            for v_group in variant_groups:
                params = {
                    'id': f"civic.variant_group:{v_group['id']}",
                    'label': v_group['name'],
                    'description': v_group['description'],
                    'type': 'variant_group'
                }
                if v_group['description'] == '':
                    del params['description']
                v_groups.append(params)
            extensions.append(Extension(
                name='variant_group',
                value=v_groups
            ).dict(exclude_none=True))
        return extensions

    def _get_variant_xrefs(self, v) -> Optional[List[str]]:
        """Return a list of xrefs for a variant.

        :param dict v: A CIViC variant record
        :return: A dictionary of xrefs
        """
        xrefs = []
        for xref in ['clinvar_entries', 'allele_registry_id',
                     'variant_aliases']:
            if xref == 'clinvar_entries':
                for clinvar_entry in v['clinvar_entries']:
                    if clinvar_entry and clinvar_entry not in ['N/A',
                                                               "NONE FOUND"]:
                        xrefs.append(f"{schemas.XrefSystem.CLINVAR.value}:"
                                     f"{clinvar_entry}")
            elif xref == 'allele_registry_id' and v['allele_registry_id']:
                xrefs.append(f"{schemas.XrefSystem.CLINGEN.value}:"
                             f"{v['allele_registry_id']}")
            elif xref == 'variant_aliases':
                dbsnp_xrefs = [item for item in v['variant_aliases']
                               if item.startswith('RS')]
                for dbsnp_xref in dbsnp_xrefs:
                    xrefs.append(f"{schemas.XrefSystem.DB_SNP.value}:"
                                 f"{dbsnp_xref.split('RS')[-1]}")
        return xrefs

    def _get_hgvs_expr(self, variant) -> Optional[List[Dict[str, str]]]:
        """Return a list of hgvs expressions for a given variant.

        :param dict variant: A CIViC variant record
        :return: A list of hgvs expressions
        """
        hgvs_expressions = list()
        for hgvs_expr in variant['hgvs_expressions']:
            if ':g.' in hgvs_expr:
                syntax = 'hgvs:genomic'
            elif ':c.' in hgvs_expr:
                syntax = 'hgvs:transcript'
            else:
                syntax = 'hgvs:protein'
            if hgvs_expr != 'N/A':
                hgvs_expressions.append(
                    Expression(syntax=syntax,
                               value=hgvs_expr).dict(exclude_none=True)
                )
        return hgvs_expressions

    def _add_gene_descriptors(self, genes) -> None:
        """Add Gene Descriptors to dict of transformations.

        :param list genes: CIViC genes
        """
        for gene in genes:
            gene_id = f"civic.gid:{gene['id']}"
            ncbigene = f"ncbigene:{gene['entrez_id']}"
            queries = [ncbigene, gene['name']] + gene['aliases']

            _, normalized_gene_id = \
                self.vicc_normalizers.normalize_gene(queries)

            if normalized_gene_id:
                gene_descriptor = GeneDescriptor(
                    id=gene_id,
                    label=gene['name'],
                    description=gene['description'] if gene['description'] else None,  # noqa: E501
                    gene_id=normalized_gene_id,
                    alternate_labels=gene['aliases'],
                    xrefs=[ncbigene]
                ).dict(exclude_none=True)
                self.gene_descriptors.append(gene_descriptor)
            else:
                logger.warning(f"Gene Normalizer unable to normalize {gene_id}"
                               f"using queries: {queries}")

    def _add_disease_descriptor(self, disease_id, record) \
            -> Optional[ValueObjectDescriptor]:
        """Add disease ID to list of valid or invalid transformations.

        :param str disease_id: The CIViC ID for the disease
        :param dict record: CIViC AID or EID
        :return: A disease descriptor
        """
        disease_descriptor = \
            self.valid_ids['disease_descriptors'].get(disease_id)
        if disease_descriptor:
            return disease_descriptor
        else:
            disease_descriptor = None
            if disease_id not in self.invalid_ids['disease_descriptors']:
                disease_descriptor = \
                    self._get_disease_descriptors(record['disease'])
                if disease_descriptor:
                    self.valid_ids['disease_descriptors'][disease_id] = \
                        disease_descriptor
                else:
                    self.invalid_ids['disease_descriptors'].append(disease_id)
            return disease_descriptor

    def _get_disease_descriptors(self, disease) \
            -> Optional[ValueObjectDescriptor]:
        """Get a disease descriptor.

        :param dict disease: A CIViC disease record
        :return: A Disease Descriptor
        """
        if not disease:
            return None

        disease_id = f"civic.did:{disease['id']}"
        display_name = disease['display_name']
        doid = disease['doid']

        if not doid:
            logger.warning(f"{disease_id} ({display_name}) has null DOID")
            queries = [display_name]
            xrefs = []
        else:
            doid = f"DOID:{disease['doid']}"
            queries = [doid, display_name]
            xrefs = [doid]

        _, normalized_disease_id = \
            self.vicc_normalizers.normalize_disease(queries)

        if not normalized_disease_id:
            logger.warning(f"Disease Normalizer unable to normalize: "
                           f"{disease_id} using queries {queries}")
            return None

        disease_descriptor = ValueObjectDescriptor(
            id=disease_id,
            type="DiseaseDescriptor",
            label=display_name,
            disease_id=normalized_disease_id,
            xrefs=xrefs if xrefs else None
        ).dict(exclude_none=True)
        return disease_descriptor

    def _add_therapy_descriptor(self, therapy_id, record)\
            -> Optional[ValueObjectDescriptor]:
        """Add therapy ID to list of valid or invalid transformations.

        :param str therapy_id: The CIViC ID for the drug
        :param dict record: CIViC AID or EID
        :return: A therapy descriptor
        """
        therapy_descriptor = \
            self.valid_ids['therapy_descriptors'].get(therapy_id)
        if therapy_descriptor:
            return therapy_descriptor
        else:
            therapy_descriptor = None
            if therapy_id not in self.invalid_ids['therapy_descriptors']:
                therapy_descriptor = \
                    self._get_therapy_descriptor(record['drugs'][0])
                if therapy_descriptor:
                    self.valid_ids['therapy_descriptors'][therapy_id] = \
                        therapy_descriptor
                else:
                    self.invalid_ids['therapy_descriptors'].append(therapy_id)
            return therapy_descriptor

    def _get_therapy_descriptor(self, drug) \
            -> Optional[ValueObjectDescriptor]:
        """Get a therapy descriptor.

        :param dict drug: A CIViC drug record
        :return: A Therapy Descriptor
        """
        therapy_id = f"civic.tid:{drug['id']}"
        label = drug['name']
        ncit_id = f"ncit:{drug['ncit_id']}"
        queries = [ncit_id, label]

        _, normalized_therapy_id = \
            self.vicc_normalizers.normalize_therapy(queries)

        if not normalized_therapy_id:
            logger.warning(f"Therapy Normalizer unable to normalize: "
                           f"using queries {ncit_id} and {label}")
            return None

        therapy_descriptor = ValueObjectDescriptor(
            id=therapy_id,
            type="TherapyDescriptor",
            label=label,
            therapy_id=normalized_therapy_id,
            alternate_labels=drug['aliases'],
            xrefs=[ncit_id]
        ).dict(exclude_none=True)
        return therapy_descriptor

    def _add_methods(self) -> None:
        """Add methods to list of transformations."""
        self.methods = [
            schemas.Method(
                id=f'method:'
                   f'{schemas.MethodID.CIVIC_EID_SOP}',
                label='Standard operating procedure for curation and clinical'
                      ' interpretation of variants in cancer',
                url='https://genomemedicine.biomedcentral.com/articles/'
                    '10.1186/s13073-019-0687-x',
                version=schemas.Date(year=2019, month=11, day=29).dict(),
                authors='Danos, A.M., Krysiak, K., Barnell, E.K. et al.'
            ).dict(exclude_none=True),
            schemas.Method(
                id=f'method:'
                   f'{schemas.MethodID.CIVIC_AID_AMP_ASCO_CAP.value}',
                label='Standards and Guidelines for the '
                      'Interpretation and Reporting of Sequence '
                      'Variants in Cancer: A Joint Consensus '
                      'Recommendation of the Association '
                      'for Molecular Pathology, American Society of '
                      'Clinical Oncology, and College of American '
                      'Pathologists',
                url='https://pubmed.ncbi.nlm.nih.gov/27993330/',
                version=schemas.Date(year=2017,
                                     month=1).dict(exclude_none=True),
                authors='Li MM, Datto M, Duncavage EJ, et al.'
            ).dict(exclude_none=True),
            schemas.Method(
                id=f'method:'
                   f'{schemas.MethodID.CIVIC_AID_ACMG.value}',
                label='Standards and guidelines for the '
                      'interpretation of sequence variants: a '
                      'joint consensus recommendation of the '
                      'American College of Medical Genetics and'
                      ' Genomics and the Association for '
                      'Molecular Pathology',
                url='https://pubmed.ncbi.nlm.nih.gov/25741868/',
                version=schemas.Date(year=2015,
                                     month=5).dict(exclude_none=True),
                authors='Richards S, Aziz N, Bale S, et al.'
            ).dict(exclude_none=True)
        ]

    def _get_eid_document(self, source) -> Optional[schemas.Document]:
        """Get an EID's document.

        :param dict source: An evidence item's source
        :return: Document for EID
        """
        source_type = source['source_type'].upper()
        if source_type in schemas.SourcePrefix.__members__:
            prefix = schemas.SourcePrefix[source_type].value
            document_id = f"{prefix}:{source['citation_id']}"
            xrefs = []
            if source['asco_abstract_id']:
                xrefs.append(f"asco.abstract:{source['asco_abstract_id']}")
            if source['pmc_id']:
                xrefs.append(f"pmc:{source['pmc_id']}")

            document = schemas.Document(
                id=document_id,
                label=source['citation'],
                description=source['name'],
                xrefs=xrefs if xrefs else None
            ).dict(exclude_none=True)
            return document
        else:
            logger.warning(f"{source_type} not in schemas.SourcePrefix")

    def _get_aid_document(self, assertion: Dict) -> List[schemas.Document]:
        """Get an AID's documents.

        :param dict assertion: A CIViC Assertion
        :return: A list of AID documents
        """
        # NCCN Guidlines
        documents = list()
        label = assertion["nccn_guideline"]
        version = assertion["nccn_guideline_version"]
        if label and version:
            doc_id = "https://www.nccn.org/professionals/physician_gls/default.aspx"  # noqa: E501
            doc_label = f"NCCN Guidelines: {label} version {version}"
            db_id = self._get_document_id(document_id=doc_id, label=doc_label)
            documents = list()
            documents.append(schemas.Document(
                id=db_id,
                document_id=doc_id,
                label=doc_label
            ).dict(exclude_none=True))

        # TODO: Check this after first pass
        # ACMG Codes
        if assertion['acmg_codes']:
            for acmg_code in assertion['acmg_codes']:
                document_id = f"acmg:{acmg_code['code']}"
                documents.append(schemas.Document(
                    id=document_id,
                    label=acmg_code['code'],
                    description=acmg_code['description']
                ).dict(exclude_none=True))

        return documents
